import warnings
from typing import Dict, List, Optional

import numpy
import onnx
import onnx.version_converter

from ._attributes import AttrGraph
from ._inline import _Inline
from ._internal_op import _InternalNode
from ._node import Node
from ._schemas import SCHEMAS
from ._scope import Scope
from ._utils import from_array
from ._var import Var


def adapt_node(
    node: Node,
    proto: onnx.NodeProto,
    source_version: int,
    target_version: int,
    var_names: Dict[Var, str],
) -> Optional[List[onnx.NodeProto]]:
    if source_version == target_version:
        return None

    try:
        # By using a dictionary we ensure that we only have a single
        # ValueInfo per (possibly repeated) input name.
        input_info = {
            var_names[var]: var.unwrap_type()._to_onnx_value_info(
                var_names[var], _traceback_name=f"adapt-input {key}"
            )
            for key, var in node.inputs.get_vars().items()
        }
        output_info = [
            var.unwrap_type()._to_onnx_value_info(
                var_names[var], _traceback_name=f"adapt-output {key}"
            )
            for key, var in node.outputs.get_vars().items()
        ]
        initializers = [
            from_array(var._value, name)
            for name, var in node.inputs.get_vars().items()
            if isinstance(var._value, numpy.ndarray)
        ]
    except ValueError:
        return None

    source_model = onnx.helper.make_model(
        onnx.helper.make_graph(
            [proto],
            "spox__singleton_adapter_graph",
            list(input_info.values()),
            output_info,
            initializers,
        ),
        opset_imports=[onnx.helper.make_operatorsetid("", source_version)],
    )
    onnx.checker.check_model(source_model, full_check=True)
    target_model = onnx.version_converter.convert_version(source_model, target_version)

    return list(target_model.graph.node)


def adapt_inline(
    node: _Inline,
    protos: List[onnx.NodeProto],
    target_opsets: Dict[str, int],
    var_names: Dict[Var, str],
    node_name: str,
) -> List[onnx.NodeProto]:
    source_version = max({v for d, v in node.opset_req if d in ("", "ai.onnx")})
    target_version = target_opsets[""]

    # convert_version fails if the inlined model does not import the default domain
    seen_domains = {prot.domain for prot in protos}
    if not seen_domains & {"", "ai.onnx"}:
        return protos
    if source_version != target_version:
        target_model = onnx.version_converter.convert_version(
            node.model, target_version
        )
        base_model = node.model
        try:
            node.model = target_model
            target_nodes = node.to_onnx(Scope.of((node, node_name), *var_names.items()))
        finally:
            node.model = base_model
        return target_nodes
    return protos


def adapt_best_effort(
    node: Node,
    protos: List[onnx.NodeProto],
    opsets: Dict[str, int],
    var_names: Dict[Var, str],
    node_names: Dict[Node, str],
) -> Optional[List[onnx.NodeProto]]:
    if isinstance(node, _Inline):
        return adapt_inline(
            node,
            protos,
            opsets,
            var_names,
            node_names[node],
        )
    if isinstance(node, _InternalNode) or len(protos) != 1:
        return None
    if any(isinstance(attr, AttrGraph) for attr in node.attrs.get_fields().values()):
        return None
    proto: onnx.NodeProto
    (proto,) = protos
    domain = proto.domain if proto.domain != "ai.onnx" else ""

    reqs = {v for d, v in node.opset_req if d == domain}
    source_version = max(reqs)
    target_version = opsets[domain]

    version_mismatch = source_version != target_version
    if version_mismatch:
        source_schema = (
            SCHEMAS.get(domain, {}).get(source_version, {}).get(proto.op_type)
        )
        target_schema = (
            SCHEMAS.get(domain, {}).get(target_version, {}).get(proto.op_type)
        )
        if source_schema is not None and target_schema is not None:
            version_mismatch = source_schema != target_schema
    if not version_mismatch:
        return None

    if proto.domain not in ("", "ai.onnx"):
        warnings.warn(
            RuntimeWarning(
                "Node adapters are only supported for the default domain (ai.onnx), "
                f"but {proto.domain!r} is at {target_version} versus requested "
                f"{source_version} of {node_names[node]}.",
            )
        )
        return None

    adapted = adapt_node(
        node,
        proto,
        source_version,
        target_version,
        var_names,
    )
    return adapted



import pytest

import spox.opset.ai.onnx.v17 as op
from spox._graph import arguments
from spox._standard import InferenceError
from spox._type_system import Tensor


def test_one_hot_inference():
    x, y, z = arguments(
        x=Tensor(int, ("N", "M")), y=Tensor(int, ()), z=Tensor(float, (2,))
    )
    assert op.one_hot(x, y, z).unwrap_tensor() == Tensor(float, ("N", "M", None))
    assert op.one_hot(x, y, z, axis=0).unwrap_tensor() == Tensor(
        float, (None, "N", "M")
    )
    assert op.one_hot(x, y, z, axis=1).unwrap_tensor() == Tensor(
        float, ("N", None, "M")
    )
    assert op.one_hot(x, y, z, axis=-1).unwrap_tensor() == Tensor(
        float, ("N", "M", None)
    )
    assert op.one_hot(x, y, z, axis=-2).unwrap_tensor() == Tensor(
        float, ("N", None, "M")
    )


def test_one_hot_inference_checks_axis_in_range():
    x, y, z = arguments(
        x=Tensor(int, ("N", "M")), y=Tensor(int, ()), z=Tensor(float, (2,))
    )
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        assert op.one_hot(x, y, z, axis=-4)
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        assert op.one_hot(x, y, z, axis=3)



import spox.opset.ai.onnx.v17 as op
from spox import Tensor, argument


def test_loop_inference():
    x, y, zs = op.loop(
        v_initial=[argument(Tensor(float, (None,))), argument(Tensor(int, ("N", 2)))],
        body=lambda i, c, a, b: [op.const(True), a, op.add(i, b), i],
    )
    assert x.type == Tensor(float, (None,))
    assert y.type == Tensor(int, ("N", 2))
    assert zs.type == Tensor(int, (None, 1))



import numpy as np
import pytest

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._standard import InferenceError
from spox._type_system import Tensor


def test_imputer_inference():
    (x,) = arguments(x=Tensor(np.int64, (None, 5, "N")))
    y = op_ml.imputer(x, imputed_value_int64s=[999], replaced_value_int64=-1)
    assert y.type == Tensor(np.int64, (None, 5, "N"))


def test_imputer_inference_with_n_feature():
    (x,) = arguments(x=Tensor(np.int64, (None, 5, "N", 3)))
    y = op_ml.imputer(
        x, imputed_value_int64s=[999, 9999, 99999], replaced_value_int64=-1
    )
    assert y.type == Tensor(np.int64, (None, 5, "N", 3))


def test_imputer_inference_with_n_feature_mismatch():
    (x,) = arguments(x=Tensor(np.int64, (None, 5, "N", 4)))
    with pytest.raises(InferenceError):
        op_ml.imputer(
            x, imputed_value_int64s=[999, 9999, 99999], replaced_value_int64=-1
        )



from dataclasses import dataclass
from typing import Iterable

import numpy
import onnx
import onnx.parser
import pytest

import spox.opset.ai.onnx.v18 as op18
import spox.opset.ai.onnx.v19 as op19
from spox import Tensor, Var, argument, build, inline
from spox._attributes import AttrInt64s
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._graph import arguments, results
from spox._node import OpType
from spox._standard import StandardNode


@pytest.fixture
def old_squeeze() -> onnx.ModelProto:
    return onnx.parser.parse_model(
        """
<
 ir_version: 8,
 opset_import: ["" : 12]
>
agraph (float[1, N] A) => (float[N] B)
{
    B = Squeeze<axes = [0]>(A)
}
"""
    )


@pytest.fixture
def old_identity() -> onnx.ModelProto:
    return onnx.parser.parse_model(
        """
<
 ir_version: 8,
 opset_import: ["" : 13]
>
agraph (float[N] A) => (float[N] B)
{
    B = Identity(A)
}
    """
    )


@pytest.fixture
def inline_old_squeeze_graph(old_squeeze):
    (data,) = arguments(
        data=Tensor(
            numpy.float32,
            (
                1,
                None,
            ),
        )
    )
    (result,) = inline(old_squeeze)(A=data).values()
    return results(final=result).with_opset(("ai.onnx", 17))


@pytest.fixture
def inline_old_identity_twice_graph(old_identity):
    (x,) = arguments(data=Tensor(numpy.float32, (None,)))
    (y,) = inline(old_identity)(A=x).values()
    (z,) = inline(old_identity)(A=y).values()
    return results(final=z).with_opset(("ai.onnx", 17))


@pytest.fixture
def old_squeeze_graph(old_squeeze):
    class Squeeze11(StandardNode):
        @dataclass
        class Attributes(BaseAttributes):
            axes: AttrInt64s

        @dataclass
        class Inputs(BaseInputs):
            data: Var

        @dataclass
        class Outputs(BaseOutputs):
            squeezed: Var

        op_type = OpType("Squeeze", "", 11)

        attrs: Attributes
        inputs: Inputs
        outputs: Outputs

    def squeeze11(_data: Var, _axes: Iterable[int]):
        return Squeeze11(
            Squeeze11.Attributes(AttrInt64s(_axes, "axes")), Squeeze11.Inputs(_data)
        ).outputs.squeezed

    (data,) = arguments(
        data=Tensor(
            numpy.float32,
            (
                1,
                None,
            ),
        )
    )
    result = squeeze11(data, [0])
    return results(final=result).with_opset(("ai.onnx", 17))


def test_adapts_inline_old_squeeze(onnx_helper, inline_old_squeeze_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            inline_old_squeeze_graph,
            "final",
            data=numpy.array([[1, 2, 3, 4]], dtype=numpy.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapts_inline_old_identity_twice(onnx_helper, inline_old_identity_twice_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            inline_old_identity_twice_graph,
            "final",
            data=numpy.array([1, 2, 3, 4], dtype=numpy.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapts_singleton_old_squeeze(onnx_helper, old_squeeze_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            old_squeeze_graph,
            "final",
            data=numpy.array([[1, 2, 3, 4]], dtype=numpy.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapt_node_with_repeating_input_names():
    a = argument(Tensor(numpy.float32, ("N",)))
    b = op18.equal(a, a)
    c = op19.identity(a)

    build({"a": a}, {"b": b, "c": c})


def test_inline_model_custom_node_only():
    """Inline a model which only consists of a custom node.

    Such models do not import from the default domain.
    """
    domain = "foo.ai"
    node = onnx.helper.make_node("FooOp", ["a"], ["b"], domain=domain)
    value_infos_input = [
        onnx.helper.make_value_info(
            "a", onnx.helper.make_tensor_type_proto(onnx.TensorProto.STRING, ("N",))
        ),
    ]
    value_infos_output = [
        onnx.helper.make_value_info(
            "b", onnx.helper.make_tensor_type_proto(onnx.TensorProto.STRING, ("N",))
        )
    ]

    model = onnx.helper.make_model(
        onnx.helper.make_graph(
            [node],
            "graph",
            value_infos_input,
            value_infos_output,
        ),
        opset_imports=[onnx.helper.make_opsetid(domain, 1)],
    )

    # Ensure that our model is valid
    onnx.checker.check_model(model, full_check=True)

    (a,) = arguments(data=Tensor(numpy.str_, ("N",)))
    (b,) = inline(model)(a).values()

    # Add another node to the model to trigger the adaption logic
    c = op18.identity(b)
    build({"a": a}, {"c": c})


@pytest.mark.skip(
    reason="Adapting custom nodes (including their subgraphs) is currently not supported"
)
def test_inline_model_custom_node_nested(old_squeeze: onnx.ModelProto):
    """A singleton custom node with a old standard node in its attribute."""
    domain = "foo.ai"

    node = onnx.helper.make_node(
        "FooOp", ["a"], ["b"], domain=domain, **{"nested_graph": old_squeeze.graph}
    )
    value_infos_input = [
        onnx.helper.make_value_info(
            "a", onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, ("N",))
        ),
    ]
    value_infos_output = [
        onnx.helper.make_value_info(
            "b", onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, ("N",))
        )
    ]

    model = onnx.helper.make_model(
        onnx.helper.make_graph(
            [node],
            "graph",
            value_infos_input,
            value_infos_output,
        ),
        opset_imports=[
            onnx.helper.make_opsetid(domain, 1),
            onnx.helper.make_opsetid("", 12),
        ],
    )

    # Ensure that our model is valid
    onnx.checker.check_model(model, full_check=True)

    (a,) = arguments(data=Tensor(numpy.float32, ("N",)))
    (b,) = inline(model)(a).values()

    # Add another node to the model to trigger the adaption logic
    c = op18.identity(b)
    build({"a": a}, {"c": c})



import dataclasses
import enum
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Iterator, Optional, Sequence, Tuple, Union

from ._attributes import Attr
from ._var import Var


@dataclass
class BaseFields:
    pass


@dataclass
class BaseAttributes(BaseFields):
    def get_fields(self) -> Dict[str, Union[None, Attr]]:
        """Return a mapping of all fields stored in this object by name."""
        return self.__dict__.copy()


class VarFieldKind(enum.Enum):
    """An enumeration of possible kinds of ONNX input/outputs fields."""

    SINGLE = 0
    OPTIONAL = 1
    VARIADIC = 2


@dataclass
class BaseVars(BaseFields):
    def __post_init__(self):
        # Check if passed fields are of the appropriate types based on field kinds
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            field_type = self._get_field_type(field)
            if field_type == VarFieldKind.SINGLE:
                if not isinstance(value, Var):
                    raise TypeError(f"Field expected Var, got: {type(value)}.")
            elif field_type == VarFieldKind.OPTIONAL:
                if value is not None and not isinstance(value, Var):
                    raise TypeError(
                        f"Optional must be Var or None, got: {type(value)}."
                    )
            elif field_type == VarFieldKind.VARIADIC:
                if not isinstance(value, Iterable):
                    raise TypeError(
                        f"Variadic field must be iterable, got '{type(value)}'."
                    )
                # Cast to tuple to avoid accidental mutation
                setattr(self, field.name, tuple(value))
                if bad := {type(var) for var in value} - {Var}:
                    raise TypeError(
                        f"Variadic field must only consist of Vars, got: {bad}."
                    )

    @classmethod
    def _get_field_type(cls, field) -> VarFieldKind:
        """Access the kind of the field (single, optional, variadic) based on its type annotation."""
        if field.type == Var:
            return VarFieldKind.SINGLE
        elif field.type == Optional[Var]:
            return VarFieldKind.OPTIONAL
        elif field.type == Sequence[Var]:
            return VarFieldKind.VARIADIC
        raise ValueError(f"Bad field type: '{field.type}'.")

    def _flatten(self) -> Iterable[Tuple[str, Optional[Var]]]:
        """Iterate over the pairs of names and values of fields in this object."""
        for key, value in self.__dict__.items():
            if value is None or isinstance(value, Var):
                yield key, value
            else:
                yield from ((f"{key}_{i}", v) for i, v in enumerate(value))

    def __iter__(self) -> Iterator[Optional[Var]]:
        """Iterate over the values of fields in this object."""
        yield from (v for _, v in self._flatten())

    def __len__(self) -> int:
        """Count the number of fields in this object (should be same as declared in the class)."""
        return sum(1 for _ in self)

    def get_vars(self) -> Dict[str, Var]:
        """Return a flat mapping by name of all the Vars in this object."""
        return {key: var for key, var in self._flatten() if var is not None}

    def get_fields(self) -> Dict[str, Union[None, Var, Sequence[Var]]]:
        """Return a mapping of all fields stored in this object by name."""
        return self.__dict__.copy()

    def _unpack_to_any(self) -> Any:
        """Unpack the stored fields into a tuple of appropriate length, typed as Any."""
        return tuple(self.__dict__.values())

    @property
    def fully_typed(self) -> bool:
        """Check if all stored variables have a concrete type."""
        return all(
            var.type is not None and var.type._is_concrete
            for var in self.get_vars().values()
        )


@dataclass
class BaseInputs(BaseVars):
    pass


@dataclass
class BaseOutputs(BaseVars):
    pass



"""Internal module implementing the low-level Graph object and functions for creating arguments and Graphs."""

import dataclasses
import itertools
from dataclasses import dataclass, replace
from typing import Callable, Dict, Iterable, List, Literal, Optional, Set, Tuple, Union

import numpy
import onnx
import onnx.shape_inference

from . import _build
from ._adapt import adapt_best_effort
from ._attributes import AttrString, AttrTensor, AttrType
from ._fields import BaseInputs
from ._internal_op import Argument, _Initializer
from ._node import Node
from ._schemas import max_opset_policy
from ._type_system import Tensor, Type
from ._utils import from_array
from ._var import Var


def arguments_dict(**kwargs: Optional[Union[Type, numpy.ndarray]]) -> Dict[str, Var]:
    """
    Parameters
    ----------
    kwargs
        Types or arrays for the newly created arguments.
        Keyword argument names are meaningful and used to name the arguments of the final graph.
        A numpy array is interpreted as an initializer (default argument value),
        and its type is used to create a respective Tensor.
    Returns
    -------
    Dict[str, Var]
        Argument Vars of given Types, named the same as kwargs.
    """
    result = {}
    for name, info in kwargs.items():
        attr_name = AttrString(value=name, name="dummy")
        if isinstance(info, Type):
            result[name] = Argument(
                Argument.Attributes(
                    name=attr_name,
                    type=AttrType(value=info, name="dummy"),
                    default=None,
                ),
                BaseInputs(),
            ).outputs.arg
        elif isinstance(info, numpy.ndarray):
            ty = Tensor(info.dtype, info.shape)
            result[name] = Argument(
                Argument.Attributes(
                    name=attr_name,
                    type=AttrType(value=ty, name="dummy"),
                    default=AttrTensor(value=info, name="dummy"),
                ),
                BaseInputs(),
            ).outputs.arg
        else:
            raise TypeError(f"Cannot construct argument from {type(info)}.")
    return result


def arguments(**kwargs: Optional[Union[Type, numpy.ndarray]]) -> Tuple[Var, ...]:
    """This function is a shorthand for a respective call to ``arguments_dict``, unpacking the Vars from the dict."""
    return tuple(arguments_dict(**kwargs).values())


def enum_arguments(
    *infos: Union[Type, numpy.ndarray], prefix: str = "in"
) -> Tuple[Var, ...]:
    """
    Convenience function for creating an enumeration of arguments, prefixed with ``prefix``.
    Calls ``arguments`` internally.

    This is a function useful for creating subgraphs, where the exact names don't really matter, only their order.
    Note that repeated use of this in the same graph may repeat names if the prefix is also the same.

    Parameters
    ----------
    infos
        Types/initializers for the created arguments.
    prefix
        String to prefix the names of created arguments with.
    Returns
    -------
    Tuple[Var, ...]
        Argument Vars as specified, in the same order as information ``infos``.
    """
    return arguments(**{f"{prefix}{i}": info for i, info in enumerate(infos)})


def initializer(arr: numpy.ndarray) -> Var:
    """
    Create a single initializer (frozen argument) with a given array value.

    This is an alternate method to creating a constant from using a dedicated Constant constructor.
    As a convention, initializers may be used for more global-scope constants.

    Parameters
    ----------
    arr
        Value of the initializer.
    Returns
    -------
        Var which is always equal to the respective value provided by `arr`.
    """
    return _Initializer(
        _Initializer.Attributes(value=AttrTensor(value=arr, name="dummy")),
        BaseInputs(),
    ).outputs.arg


@dataclass(frozen=True, eq=False)
class Graph:
    """
    Represents an abstraction for a wrapped up ONNX computation graph,
    that can be built into ONNX GraphProto & ModelProto.

    Should be constructed only with the ``results`` functions.

    Use the methods ``rename``, ``doc``, ``with_arguments`` to set additional data for the graph.
    These methods return a new instance of Graph with a respective private attribute set.

    Note that to not only fix results (which a Graph is constructed with), but also arguments, ``with_arguments``
    should be used.

    Note: building a Graph is cached, so changing it in-place without the setters will invalidate the build.
    """

    _results: Dict[str, Var]
    _name: Optional[str] = None
    _doc_string: Optional[str] = None
    _arguments: Optional[Tuple[Var, ...]] = None
    _extra_opset_req: Optional[Set[Tuple[str, int]]] = None
    _constructor: Optional[Callable[..., Iterable[Var]]] = None
    _build_result: "_build.Cached[_build.BuildResult]" = dataclasses.field(
        default_factory=_build.Cached
    )

    def __repr__(self):
        name_repr = self._name if self._name is not None else "?"
        args_repr = (
            f"{', '.join(str(a) for a in self._arguments)}"
            if self._arguments is not None
            else "..."
        )
        res_repr = f"{', '.join(f'{k}: {a}' for k, a in self._results.items())}"
        comments: List[str] = []
        if self._doc_string is not None:
            comments.append(f'"{self._doc_string[:10]}..."')
        if self._extra_opset_req is not None:
            comments.append(f"+{len(self._extra_opset_req)} opset req")
        return f"<Graph '{name_repr}' ({args_repr}) -> ({res_repr}){': ' if comments else ''}{', '.join(comments)}>"

    def __post_init__(self):
        if any(not isinstance(var, Var) for var in self._results.values()):
            seen_types = {type(obj) for obj in self._results.values()}
            raise TypeError(f"Graph results must be Vars, not {seen_types - {Var}}.")
        if self._arguments is not None and any(
            not isinstance(var, Var) for var in self._arguments
        ):
            seen_types = {type(obj) for obj in self._arguments}
            raise TypeError(f"Build outputs must be Vars, not {seen_types - {Var}}.")

    def with_name(self, name: str) -> "Graph":
        """Return a Graph with its name set to ``name``."""
        return replace(self, _name=name)

    def with_doc(self, doc_string: str) -> "Graph":
        """Return a Graph with its doc string set to ``doc``."""
        return replace(self, _doc_string=doc_string)

    def with_arguments(self, *args: Var) -> "Graph":
        """
        Return a Graph with given Vars marked as exactly its arguments.
        A useful idiom is ``results(...).with_arguments(...)`` when you want to specify both results and arguments.
        """
        return replace(self, _arguments=args)

    def with_opset(self, *args: Tuple[str, int]) -> "Graph":
        """
        Add the given minimum opset requirements to the graph.
        Useful when the graph is using legacy nodes, but Spox should attempt to convert them to a required version.
        """
        extra_opset_req = set(args)
        if self._extra_opset_req is not None:
            extra_opset_req |= self._extra_opset_req
        return replace(self, _extra_opset_req=extra_opset_req)

    def _with_constructor(self, fun: Callable[..., Iterable[Var]]) -> "Graph":
        """Assign a constructor that constructed this Graph given ``self.requested_arguments``."""
        return replace(self, _constructor=fun)

    def _reconstruct(self, *args: Var) -> "Graph":
        assert self._constructor is not None
        return (
            results(**dict(zip(self._results, self._constructor(*args))))
            .with_arguments(*args)
            ._with_constructor(self._constructor)
        )

    def _inject_build_result(self, what: "_build.BuildResult") -> "Graph":
        """
        Internal function used to build a Graph with a custom build result.
        Used when building subgraphs to have further control over the build state.
        """
        return replace(self, _build_result=_build.Cached(what))

    @property
    def requested_arguments(self) -> Optional[Iterable[Var]]:
        """Arguments requested by this Graph (for building) - ``None`` if unspecified."""
        return self._arguments

    @property
    def requested_results(self) -> Dict[str, Var]:
        """Results (named) requested by this Graph (for building)."""
        return self._results

    def get_arguments(self) -> Dict[str, Var]:
        """
        Get the effective named arguments (after build) of this Graph.

        May be expensive, as it has to build Use ``requested_arguments`` for a cheaper variant that may be sufficient.
        """
        return {
            self._get_build_result().scope.var[var]: var
            for var in self._get_build_result().arguments
        }

    def get_results(self) -> Dict[str, Var]:
        """
        Get the effective named results (after build) of this Graph.

        May be expensive, as it has to build. Use ``requested_results`` for a cheaper variant that may be sufficient.
        """
        return {
            self._get_build_result().scope.var[var]: var
            for var in self._get_build_result().results
        }

    def get_opsets(self) -> Dict[str, int]:
        """
        Get the effective opsets used by this Graph. The used policy for mixed versions is maximum-requested.

        May be expensive, as it has to build.
        """
        return max_opset_policy(self._get_opset_req())

    def _get_build_result(self) -> "_build.BuildResult":
        """Internal function for getting (with cache) the build result structure for this Graph."""
        if self._build_result._value is None:
            self._build_result.value = _build.Builder(self).build_main()
        return self._build_result.value

    def _get_opset_req(self) -> Set[Tuple[str, int]]:
        """Internal function for accessing the opset requirements, including extras requested by the Graph itself."""
        return self._get_build_result().opset_req | (
            self._extra_opset_req if self._extra_opset_req is not None else set()
        )

    def _get_initializers_by_name(self) -> Dict[str, numpy.ndarray]:
        """Internal function for accessing the initializers by name in the build."""
        return {
            self._get_build_result().scope.var[var]: init
            for var, init in self._get_build_result().initializers.items()
        }

    def get_adapted_nodes(self) -> Dict[Node, Tuple[onnx.NodeProto, ...]]:
        """
        Do a best-effort at generating NodeProtos of consistent versions, matching ``self.opsets``.
        In essence, the policy is to upgrade to the highest used version.
        This does not attempt to fix too complicated nodes, but should work for inline models and simple single nodes.

        Note that onnx.version_converter only implements conversion for the default domain.
        """
        nodes = self._get_build_result().nodes
        consistent_nodes = nodes.copy()
        for node, protos in nodes.items():
            best_effort = adapt_best_effort(
                node,
                list(protos),
                self.get_opsets(),
                self._get_build_result().scope.var.name_of,
                self._get_build_result().scope.node.name_of,
            )
            consistent_nodes[node] = (
                tuple(best_effort) if best_effort is not None else protos
            )

        return consistent_nodes

    def to_onnx(self, *, concrete: bool = False) -> onnx.GraphProto:
        """Perform the Spox build process, gathering arguments,
        results, nodes and other information.

        - Saves type information for arguments & results.
        - Sets the name of the graph, with defaults if it is not set
          or if it is a subgraphs
        - Saves initializers.
        - Sets the docstring if one is set.

        Note
        ----

        This function returns a `onnx.GraphProto` object! If
        you want to serialize this model in a way that is readable by
        the `onnxruntime`, you should use the `to_onnx_model` method
        instead!

        Returns
        -------
        onnx.GraphProto
            Translation of this Graph into an ONNX GraphProto object.

        """
        if not self.get_results():
            raise ValueError("Attempt to build graph without results.")

        argument_info = [
            var.unwrap_type()._to_onnx_value_info(
                name, concrete=concrete, _traceback_name=f"argument {name} ({var})"
            )
            for name, var in self.get_arguments().items()
        ]
        result_info = [
            var.unwrap_type()._to_onnx_value_info(
                name, concrete=concrete, _traceback_name=f"result {name} ({var})"
            )
            for name, var in self.get_results().items()
        ]

        if self._name:
            name = self._name
        else:
            name = "spox_graph"

        initializer_tensors = [
            from_array(arr, name)
            for name, arr in self._get_initializers_by_name().items()
        ]

        node_protos = itertools.chain.from_iterable(self.get_adapted_nodes().values())
        return onnx.helper.make_graph(
            list(node_protos),
            name,
            argument_info,
            result_info,
            initializer_tensors,
            self._doc_string,
        )

    def to_onnx_model(
        self,
        *,
        producer_name: str = "spox",
        model_doc_string: str = "",
        infer_shapes: bool = False,
        check_model: Union[Literal[0], Literal[1], Literal[2]] = 1,
        ir_version=8,
        concrete: bool = True,
    ) -> onnx.ModelProto:
        """
        Internally, this function first obtains a GraphProto from ``.to_onnx()``. Additionally:

        - Function definitions are collected and built into FunctionProtos.
        - Opset requirements are collected (consistency policy is to use the highest version).
            ONNX only allows one version of each domain per model, so some attempt at conversion of nodes is made.
        - Checks are performed, at the level described by the respective arguments.

        Parameters
        ----------
        producer_name
            Value of the ONNX ModelProto producer name field.
        model_doc_string
            Doc string for the ONNX ModelProto.
        infer_shapes
            If the value is True, the model is passed through `onnx.shape_inference.infer_shapes`.
        check_model
            If the value is at least 1 (default), `onnx.checker.check_model` is executed on the model.
            If it is 2, `full_check` of the `check_model` call is set to `True` (e.g. tests against shape inference).
        concrete
            Whether to raise for non-concrete value infos (like missing shape information).
        Returns
        -------
            Translation of this Graph into an ONNX ModelProto object.
        """

        opsets = self.get_opsets()
        if not opsets:
            raise RuntimeError(
                "ONNX often does not properly handle graphs which are empty, "
                "and this one seems to contain no opset imports (only internal nodes?). "
                "Consider adding an Identity operator if you are just copying arguments."
            )

        opset_req: List[tuple[str, int]] = list(opsets.items())  # type: ignore
        function_protos: Dict[Tuple[str, str], onnx.FunctionProto] = {}
        for fun in self._get_build_result().functions:
            proto = fun.to_onnx_function(extra_opset_req=opset_req)
            if proto is None:
                continue
            key = (proto.domain, proto.name)
            if key in function_protos and proto != function_protos[key]:
                raise RuntimeError(
                    f"Built dependency function {proto.domain}:{proto.name} has two different definitions. "
                    f"Was its implementation non-deterministic or is there a naming collision?"
                )
            function_protos[key] = proto

        model = onnx.helper.make_model(
            self.to_onnx(concrete=concrete),
            producer_name=producer_name,
            doc_string=model_doc_string,
            functions=list(function_protos.values()),
            opset_imports=[
                onnx.helper.make_operatorsetid(domain, version)
                for domain, version in opsets.items()
            ],
            ir_version=ir_version,
        )

        if infer_shapes:
            model = onnx.shape_inference.infer_shapes(model)
        if check_model:
            onnx.checker.check_model(model, full_check=check_model >= 2)
        return model


def results(**kwargs: Var) -> Graph:
    """
    Use this function to construct a ``Graph`` object.

    Parameters
    ----------
    kwargs
        Vars to be marked as results in the created Graph.
    Returns
    -------
    Graph
        Graph with the results given in `kwargs`, in the same order. Keys are used as names for the results.
    """
    return Graph(kwargs)


def enum_results(*vars: Var, prefix="out") -> Graph:
    """
    Use this function to construct a ``Graph`` object, whenever the exact names are not important.
    Useful when creating subgraphs.

    Parameters
    ----------
    vars
        Vars to be marked as results.
    prefix
        String to prefix the names of created results with.
    Returns
    -------
        Graph with the results given in `vars`, in the same order.
        Names are the `prefix` with an enumeration index at the end.
    """
    return results(**{f"{prefix}{i}": var for i, var in enumerate(vars)})


def subgraph(types: Iterable[Type], fun: Callable[..., Iterable[Var]]) -> Graph:
    """
    Convenience function for creating a subgraph, for use in an operator like If or Loop.
    However, for those operators one may prefer to use alternative constructors like ``xif`` or ``xloop``
    (which use this function internally).

    Parameters
    ----------
    types
        A list of argument types for the subgraph.
    fun
        A function taking as many Var arguments as the length of `types`, and returning the results of the subgraph.
    Returns
    -------
    Graph
        Graph with results based on the return value of `fun`.
    """
    if not (
        isinstance(types, Iterable) and all(isinstance(typ, Type) for typ in types)
    ):
        raise TypeError("Subgraph input types must be an Iterable of Type.")
    ins = enum_arguments(*types)
    for var in ins:
        var._rename(None)
    if not callable(fun):
        raise TypeError("Subgraph callback must be callable.")
    outs = fun(*ins)
    if not (isinstance(outs, Iterable) and all(isinstance(out, Var) for out in outs)):
        raise TypeError("Subgraph result must be an Iterable of Var.")
    return enum_results(*outs).with_arguments(*ins)._with_constructor(fun)



import re
from typing import Any

import numpy
import pytest

import spox.opset.ai.onnx.v17 as op
from spox._attributes import AttrTensors
from spox._exceptions import InferenceError
from spox._graph import arguments
from spox._type_system import Tensor


def test_basic_inference():
    a, b = arguments(a=Tensor(numpy.float32, ("N",)), b=Tensor(numpy.float32, (2, "N")))
    assert op.add(a, b).type == Tensor(numpy.float32, (2, "N"))


def test_variadic_input_inference():
    typ = Tensor(numpy.float32, ("N",))
    a, b, c = arguments(a=typ, b=typ, c=typ)
    assert op.max([a, b, c]).type == typ


def test_variadic_output_inference():
    (x,) = arguments(x=Tensor(numpy.float32, (3, "N")))
    x1, x2, x3 = op.split(x, op.const([1, 1, 1]), outputs_count=3)
    assert x1.type == x2.type == x3.type == Tensor(numpy.float32, (1, "N"))


def test_optional_input_inference():
    (x,) = arguments(x=Tensor(numpy.float32, ("N",)))
    assert op.clip(x, max=op.constant(value_float=1.0)).type == Tensor(
        numpy.float32, ("N",)
    )
    assert op.clip(x, min=op.constant(value_float=0.0)).type == Tensor(
        numpy.float32, ("N",)
    )
    assert op.clip(
        x, min=op.constant(value_float=0.0), max=op.constant(value_float=1.0)
    ).type == Tensor(numpy.float32, ("N",))


def test_function_body_inference():
    a, b = arguments(a=Tensor(numpy.float32, ("N",)), b=Tensor(numpy.float32, ("N",)))
    assert op.greater_or_equal(a, b).type == Tensor(numpy.bool_, ("N",))


def test_inference_fails():
    a, b = arguments(a=Tensor(numpy.float32, (2,)), b=Tensor(numpy.float32, (3,)))
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        op.add(a, b)


def test_inference_validation_fails():
    a, b = arguments(a=Tensor(numpy.float32, (2,)), b=Tensor(numpy.float64, (2,)))
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        op.add(a, b)


def test_multiple_outputs():
    x, k = arguments(
        a=Tensor(numpy.float32, ("N", "M", "K")), b=Tensor(numpy.int64, (1,))
    )
    values, indices = op.top_k(x, k)
    assert values.unwrap_type()._subtype(Tensor(numpy.float32, ("N", "M", None)))
    assert indices.unwrap_type()._subtype(Tensor(numpy.int64, ("N", "M", None)))


@pytest.mark.parametrize(
    "key,values,match",
    [
        (
            "value_ints",
            ["a"],
            "Unable to instantiate `AttrInt64s` from items of type(s) `str`.",
        ),
        (
            "value_strings",
            [1, 2.3],
            "Unable to instantiate `AttrStrings` from items of type(s) `float, int`.",
        ),
        (
            "value_floats",
            ["a"],
            "Unable to instantiate `AttrFloat32s` from items of type(s) `str`.",
        ),
        (
            "value_int",
            "a",
            "Unable to instantiate `AttrInt64` with value of type `str`.",
        ),
    ],
)
def test_passing_wrong_type(key: str, values: Any, match: str):
    with pytest.raises(TypeError, match=re.escape(match)):
        op.constant(**{key: values})


def test_passing_wrong_type_tensors():
    with pytest.raises(
        TypeError,
        match=re.escape(
            "Unable to instantiate `AttrTensors` from items of type(s) `int`."
        ),
    ):
        AttrTensors([1], "dummy")  # type: ignore



from typing import Optional

import numpy as np
import numpy.typing as npt
import onnx
from onnx import TensorProto


def tensor_type_to_dtype(ttype: int) -> np.dtype:
    """Convert integer tensor types to ``numpy.dtype`` objects."""
    if ttype == onnx.TensorProto.STRING:
        return np.dtype(str)  # Spox uses the str datatype for strings, not object
    return onnx.helper.tensor_dtype_to_np_dtype(ttype)


def dtype_to_tensor_type(dtype_like: npt.DTypeLike) -> int:
    """Convert numpy data types into integer tensor types.

    Raises
    ------
    TypeError:
        If ``dtype_like`` has no corresponding tensor type in the ONNX standard.
    """
    err_msg = f"{dtype_like} is not a valid ONNX tensor element type."
    if dtype_like is None:  # numpy would default to float64
        raise TypeError(err_msg)
    # normalize in the case of aliases like ``long`` which are missing in the lookup
    dtype = np.dtype(np.dtype(dtype_like).type)
    if dtype == np.dtype(object):
        raise TypeError(
            "`np.dtype('object')` is not supported as a tensor element type. "
            "Hint: Spox uses `np.dtype('str')` for the string datatype."
        )
    elif dtype == np.dtype(str):
        return onnx.TensorProto.STRING
    try:
        return onnx.helper.np_dtype_to_tensor_dtype(dtype)
    except KeyError:
        raise TypeError(err_msg)


def from_array(arr: np.ndarray, name: Optional[str] = None) -> TensorProto:
    """Convert the given ``numpy.array`` into an ``onnx.TensorProto``.

    As it may be useful to name the TensorProto (e.g. in
    initializers), there is a ``name`` parameter.

    This function differs from ``onnx.numpy_helper.from_array`` by not
    using the ``raw_data`` field.
    """
    cast_to_bytes = False
    if arr.dtype.type in [np.str_, np.object_]:
        cast_to_bytes = True
    return onnx.helper.make_tensor(
        name=name or "",
        data_type=dtype_to_tensor_type(arr.dtype),
        dims=arr.shape,
        # make_tensor fails on scalars. We fix it by calling flatten
        vals=(
            np.char.encode(arr, encoding="utf-8") if cast_to_bytes else arr
        ).flatten(),
        raw=False,
    )



import typing
from dataclasses import dataclass
from typing import TypeVar

import numpy as np
import numpy.typing as npt
import onnx

from ._shape import Shape, SimpleShape
from ._utils import dtype_to_tensor_type, tensor_type_to_dtype

T = TypeVar("T")
S = TypeVar("S")


@dataclass(frozen=True)
class Type:
    """Base class for classes representing ONNX types in Spox."""

    @classmethod
    def _from_onnx(cls, proto: onnx.TypeProto) -> "Type":
        """
        Parameters
        ----------
        proto
            Protobuf object to translate from.

        Returns
        -------
        Type
            Respective subtype of Type representing the ONNX type in
            the protobuf object.

        Raises
        ------
        ValueError
            If the passed protobuf does not contain any of the
            expected fields (tensor, sequence, optional).
        """
        if proto.HasField("tensor_type"):
            return Tensor(
                tensor_type_to_dtype(proto.tensor_type.elem_type),
                Shape.from_onnx(proto.tensor_type.shape).to_simple()
                if proto.tensor_type.HasField("shape")
                else None,
            )
        elif proto.HasField("sequence_type"):
            return Sequence(Type._from_onnx(proto.sequence_type.elem_type))
        elif proto.HasField("optional_type"):
            return Optional(Type._from_onnx(proto.optional_type.elem_type))
        raise ValueError(
            f"Cannot get Type from invalid protobuf (not tensor, sequence or optional): {proto}"
        )

    def _assert_concrete(self, *, _traceback_name: str = "?"):
        """Function used by the build process to check if a type is
        well-specified (e.g. Tensor shape is defined).

        Inheritors of ``Type`` should throw if they do not specify enough
        information to be accepted as Model input/outputs.
        """
        return self

    @property
    def _is_concrete(self) -> bool:
        try:
            self._assert_concrete()
        except Exception:
            return False
        else:
            return True

    def unwrap_tensor(self) -> "Tensor":
        """
        Return ``self``, unless this Type is not a Tensor.

        Raises
        ------
        TypeError
            If the type isn't a Tensor.

        """
        if not isinstance(self, Tensor):
            raise TypeError(f"Cannot unwrap requested Tensor type from {self}")
        return self

    def unwrap_sequence(self) -> "Sequence":
        """
        Return ``self``, unless this Type is not a Sequence.

        Raises
        ------
        TypeError
            If the type isn't a Sequence.
        """
        if not isinstance(self, Sequence):
            raise TypeError(f"Cannot unwrap requested Sequence type from {self}")
        return self

    def unwrap_optional(self) -> "Optional":
        """
        Return ``self``, unless this Type is not an Optional.

        Raises
        ------
        TypeError
            If the type isn't an Optional.

        """
        if not isinstance(self, Optional):
            raise TypeError(f"Cannot unwrap requested Optional type from {self}")
        return self

    def _to_onnx(self) -> onnx.TypeProto:
        """Translate ``self`` into an ONNX TypeProto."""
        raise TypeError(
            f"Cannot generate ONNX TypeProto for {self} (not implemented or bad type)."
        )

    def _to_onnx_value_info(
        self,
        name: str,
        doc_string: str = "",
        *,
        concrete: bool = False,
        _traceback_name: str = "?",
    ) -> onnx.ValueInfoProto:
        """Translation of ``self`` into an ONNX ValueInfoProto"""
        if concrete:
            self._assert_concrete(_traceback_name=_traceback_name)
        return onnx.helper.make_value_info(
            name,
            self._to_onnx(),
            doc_string,
        )

    def _subtype(self, other: "Type") -> bool:
        """
        Compare Types for membership.
        An Unknown field (like an unspecified Tensor shape) is treated as "any" in this comparison.
        """
        if not isinstance(other, Type):
            return NotImplemented
        return other == Type() or self == other


@dataclass(frozen=True)
class Tensor(Type):
    """
    Represents a ``Tensor`` of given ``dtype`` and ``shape``.

    The ``dtype`` describes the element type of the ``Tensor``.
    It must correspond to an allowed ONNX tensor element type.

    A shape is denoted with a tuple (simplified) format, where each
    element describes the respective axis.  The types used may be:

    - An ``int`` denoting a statically known length
    - A ``str`` denoting a named runtime-determined length
    - ``None`` representing any length.

    The ``shape`` may also be ``None`` if the rank of the ``Tensor``
    is unknown.

    If you want to specify that dimensions will be equal, you can use
    the same parameter strings.  However, this is not very strictly
    enforced.
    """

    _elem_type: typing.Type[np.generic]
    _shape: Shape

    def __init__(
        self,
        dtype: npt.DTypeLike,
        shape: SimpleShape = None,
    ):
        """Create a ``Tensor``.

        Raises
        ------
        TypeError
            If the passed ``elem_type`` does not correspond to one of
            the following numpy scalar types: ``numpy.bool_``,
            ``numpy.complex128``, ``numpy.complex64``,
            ``numpy.float16``, ``numpy.float32``, ``numpy.float64``,
            ``numpy.int16``, ``numpy.int32``, ``numpy.int64``,
            ``numpy.int8``, ``numpy.str_``, ``numpy.uint16``,
            ``numpy.uint32``, ``numpy.uint64``, ``numpy.uint8``.

        """
        # Try converting to a tensor type. If it fails, we allow the
        # exception to bubble up.
        dtype_to_tensor_type(dtype)
        rich_shape = Shape.from_simple(shape)
        object.__setattr__(self, "_elem_type", np.dtype(dtype).type)
        object.__setattr__(self, "_shape", rich_shape)

    @property
    def dtype(self) -> np.dtype:
        """Data type of this tensor."""
        return np.dtype(self._elem_type)

    @property
    def shape(self) -> SimpleShape:
        """
        Return the shape of this tensor in a simplified/tuple format
        (as used by the ``onnx`` module).  Each element of the
        ``SimpleShape`` tuple denotes information about the respective
        axis.

        Returns
        -------
        SimpleShape
            The shape of this Tensor. If it is unknown, ``None`` is returned,
            otherwise it is a tuple describing each dimension.

        """
        return self._shape.to_simple()

    def _to_onnx(self) -> onnx.TypeProto:
        return onnx.helper.make_tensor_type_proto(
            dtype_to_tensor_type(self._elem_type), self.shape
        )

    def _assert_concrete(self, *, _traceback_name: str = "?"):
        if self.shape is None:
            raise ValueError(
                f"Tensor {self} does not specify the shape -- in {_traceback_name}."
            )
        return self

    def __repr__(self):
        return f"{type(self).__name__}(dtype={self._elem_type.__name__}, shape={self.shape})"

    def __str__(self):
        dims = self.shape
        dims_repr = (
            "".join(f"[{dim if dim is not None else '?'}]" for dim in dims)
            if dims is not None
            else "[...]"
        )
        return f"{self._elem_type.__name__.rstrip('_')}" + dims_repr

    def _subtype(self, other: Type) -> bool:
        if not isinstance(other, Type):
            return NotImplemented
        if other == Type() or self == other:
            return True
        if not isinstance(other, Tensor):
            return False
        return (
            issubclass(self._elem_type, other._elem_type)
            and self._shape <= other._shape
        )


@dataclass(frozen=True)
class Sequence(Type):
    """Ordered collection of elements that are of homogeneous types."""

    elem_type: Type

    def _to_onnx(self) -> onnx.TypeProto:
        return onnx.helper.make_sequence_type_proto(self.elem_type._to_onnx())

    def __repr__(self):
        return f"{type(self).__name__}(elem_type={self.elem_type!r}"

    def __str__(self):
        return f"[{self.elem_type}]"

    def _subtype(self, other: Type) -> bool:
        if not isinstance(other, Type):
            return NotImplemented
        if other == Type() or self == other:
            return True
        if not isinstance(other, Sequence):
            return False
        return self.elem_type._subtype(other.elem_type)


@dataclass(frozen=True)
class Optional(Type):
    """
    Wrapper that may contain an element of :class:`~spox.Tensor` or
    :class:`~spox.Sequence` type, or may be empty (containing none).

    """

    elem_type: Type

    def _to_onnx(self) -> onnx.TypeProto:
        return onnx.helper.make_optional_type_proto(self.elem_type._to_onnx())

    def __repr__(self):
        return f"{type(self).__name__}(elem_type={self.elem_type!r}"

    def __str__(self):
        return f"{self.elem_type}?"

    def _subtype(self, other: Type) -> bool:
        if not isinstance(other, Type):
            return NotImplemented
        if other == Type() or self == other:
            return True
        if not isinstance(other, Optional):
            return False
        return self.elem_type._subtype(other.elem_type)



from typing import List, Tuple

import numpy
import onnx
import pytest

from spox._shape import Shape
from spox._type_system import Optional, Sequence, Tensor, Type
from spox._utils import dtype_to_tensor_type, tensor_type_to_dtype


def tensor_type_proto(elem_type, shape):
    return onnx.helper.make_tensor_type_proto(dtype_to_tensor_type(elem_type), shape)


def tensor_shape_proto(shape):
    proto = tensor_type_proto(numpy.float32, shape)
    return proto.tensor_type.shape if proto.tensor_type.HasField("shape") else None


@pytest.fixture
def tensor_elem_type_pairs():
    return [
        (numpy.float32, onnx.TensorProto.FLOAT),
        (numpy.float64, onnx.TensorProto.DOUBLE),
        (numpy.int32, onnx.TensorProto.INT32),
        (numpy.int64, onnx.TensorProto.INT64),
        (numpy.bool_, onnx.TensorProto.BOOL),
        (numpy.str_, onnx.TensorProto.STRING),
    ]


@pytest.fixture
def tensor_shape_pairs():
    return [
        (None, tensor_shape_proto(None)),
        ((), tensor_shape_proto(())),
    ]


@pytest.fixture
def type_pairs() -> List[Tuple[Type, onnx.TypeProto]]:
    tensor_f32 = tensor_type_proto(numpy.float32, None)
    seq_tensor_f32 = onnx.helper.make_sequence_type_proto(tensor_f32)
    opt_seq_tensor_f32 = onnx.helper.make_optional_type_proto(seq_tensor_f32)
    return [
        (Tensor(numpy.float32), tensor_f32),
        (Sequence(Tensor(numpy.float32)), seq_tensor_f32),
        (Optional(Sequence(Tensor(numpy.float32))), opt_seq_tensor_f32),
        (Tensor(numpy.int32, ("N", 1, 2)), tensor_type_proto(numpy.int32, ("N", 1, 2))),
        (Tensor(numpy.bool_, ()), tensor_type_proto(numpy.bool_, ())),
        (Tensor(numpy.str_, (None,)), tensor_type_proto(numpy.str_, (None,))),
    ]


def test_tensor_elem_type_to_onnx(tensor_elem_type_pairs):
    for first, second in tensor_elem_type_pairs:
        assert dtype_to_tensor_type(first) == second


def test_tensor_elem_type_from_onnx(tensor_elem_type_pairs):
    for first, second in tensor_elem_type_pairs:
        assert tensor_type_to_dtype(second) == first


def test_scalar_is_not_unknown_shape():
    assert Tensor(numpy.float32)._to_onnx() != Tensor(numpy.float32, ())._to_onnx()
    assert Shape.from_simple(None) != Shape.from_simple(())
    assert Shape.from_simple(None).to_onnx() != Shape.from_simple(()).to_onnx()


def test_tensor_shape_to_onnx(tensor_shape_pairs):
    for first, second in tensor_shape_pairs:
        assert Shape.from_simple(first).to_onnx() == second


def test_tensor_shape_from_onnx(tensor_shape_pairs):
    for first, second in tensor_shape_pairs:
        assert Shape.from_onnx(second) == Shape.from_simple(first)


def test_type_to_onnx(type_pairs):
    for first, second in type_pairs:
        assert first._to_onnx() == second


def test_type_from_onnx(type_pairs):
    for first, second in type_pairs:
        assert Type._from_onnx(second) == first



import spox.opset.ai.onnx.ml.v3 as ml3
import spox.opset.ai.onnx.v17 as op17
import spox.opset.ai.onnx.v18 as op18
import spox.opset.ai.onnx.v19 as op19


def test_ai_onnx_v17():
    op17.add(op17.constant(value_int=2), op17.constant(value_int=2))


def test_ai_onnx_v18():
    op18.bitwise_and(op17.constant(value_int=7), op17.constant(value_int=10))
    assert len(op18.split(op18.const([1, 2, 3]), num_outputs=3)) == 3


def test_ai_onnx_v19():
    op19.equal(op19.constant(value_string="abc"), op19.constant(value_string="abc"))


def test_ai_onnx_ml_v3():
    ml3.label_encoder(
        op17.constant(value_ints=[1, 2, 3]),
        keys_int64s=[0, 1, 2],
        values_strings=["a", "b", "c"],
        default_string="?",
    )



"""
Data representation for Tensor shapes. Largely attempts to mimic semantics of ONNX shapes.
Shapes have 3 representations:

- ONNX, composed of the member of TensorProto: TensorShapeProto with members TensorShapeProto.Dimension
- Simplified, also used by ONNX helper and convenient for printing and using in Python. This is SimpleShape in the code,
  and it's a tuple of mixed types (int, str, None).
- Class, implemented by Natural and Shape, which also exposes some useful methods for conversion.

Shapes may be unknown (when the rank is unknown, unknown dimensions are OK), but this cannot be used in many contexts.
Unknown shapes may cause warnings or errors to be raised.
"""

import abc
import typing
from dataclasses import dataclass
from typing import Optional, Tuple, TypeVar, Union

import onnx

SimpleShapeElem = Union[str, int, None]
SimpleShape = Optional[Tuple[SimpleShapeElem, ...]]


class ShapeError(TypeError):
    pass


@dataclass(frozen=True)
class Natural(abc.ABC):
    """Family of types used for storing known or unknown natural numbers, primarily for tensor shapes."""

    @classmethod
    def from_simple(cls, value: SimpleShapeElem) -> "Natural":
        """Translate into Natural from simplified representation."""
        if isinstance(value, int):
            return Constant(value)
        elif isinstance(value, str):
            return Unknown(value)
        elif value is None:
            return Unknown()
        raise TypeError(
            f"Cannot translate to Natural from simple shape element {value}."
        )

    @classmethod
    def simple_from_onnx(
        cls, proto: onnx.TensorShapeProto.Dimension
    ) -> SimpleShapeElem:
        """Translate an ONNX dimension element into the simplified representation."""
        if proto.HasField("dim_value"):
            return int(proto.dim_value)
        elif proto.HasField("dim_param"):
            return str(proto.dim_param)
        else:
            return None

    @classmethod
    def simple_to_onnx(cls, value: SimpleShapeElem) -> onnx.TensorShapeProto.Dimension:
        """Translate simplified representation element into an ONNX one."""
        if isinstance(value, int):
            return onnx.TensorShapeProto.Dimension(dim_value=value)
        if isinstance(value, str):
            return onnx.TensorShapeProto.Dimension(dim_param=value)
        elif value is None:
            return onnx.TensorShapeProto.Dimension()
        raise TypeError(f"Cannot translate to ONNX from simple shape element {value}.")

    def to_onnx(self) -> onnx.TensorShapeProto.Dimension:
        """Translate self into ONNX representation."""
        raise self.simple_to_onnx(self.to_simple())

    @classmethod
    def from_onnx(cls, proto: onnx.TensorShapeProto.Dimension) -> "Natural":
        """Translate Natural from ONNX dimension element."""
        return cls.from_simple(cls.simple_from_onnx(proto))

    def to_simple(self) -> SimpleShapeElem:
        """Translate self into simplified representation."""
        raise NotImplementedError(f"Cannot translate {self} to simple shape element.")

    def __le__(self, other: "Natural") -> bool:
        """Shape dimension membership comparison, with Unknown serving as an "any" quantifier."""
        if not isinstance(other, Natural):
            return NotImplemented
        return other == Unknown()


@dataclass(frozen=True)
class Unknown(Natural):
    """Represents an unknown natural number. Unknown passes any dimension constraint."""

    label: str = ""

    def to_simple(self) -> Union[str, None]:
        return None if not self.label else self.label

    def __le__(self, other: Natural) -> bool:
        if not isinstance(other, Natural):
            return NotImplemented
        return True


@dataclass(frozen=True)
class Constant(Natural):
    """Represents a given constant natural number."""

    n: int

    def to_simple(self) -> int:
        return self.n

    def __le__(self, other: Natural) -> bool:
        if not isinstance(other, Natural):
            return NotImplemented
        return isinstance(other, Unknown) or self == other


ShapeT = TypeVar("ShapeT", bound="Shape")


@dataclass(frozen=True)
class Shape:
    """Type representing a static Tensor shape."""

    dims: Optional[Tuple[Natural, ...]]

    def __bool__(self):
        return self.dims is not None

    @classmethod
    def from_simple(cls: typing.Type[ShapeT], shape: SimpleShape) -> ShapeT:
        """Translate into a Shape from the simplified representation."""
        return cls(
            tuple(Natural.from_simple(v) for v in shape) if shape is not None else None
        )

    @classmethod
    def from_onnx(
        cls: typing.Type[ShapeT], proto: Optional[onnx.TensorShapeProto]
    ) -> ShapeT:
        """Translate into a Shape from ONNX shape."""
        return (
            cls(tuple(Natural.from_onnx(dim) for dim in proto.dim))
            if proto is not None
            else cls(None)
        )

    def to_simple(self) -> SimpleShape:
        """Translate into the simplified representation."""
        return (
            tuple(v.to_simple() for v in self.dims) if self.dims is not None else None
        )

    def to_onnx(self) -> Optional[onnx.TensorShapeProto]:
        """Translate into the ONNX representation."""
        if self.dims is None:
            return None
        proto = onnx.TensorShapeProto(dim=(v.to_onnx() for v in self.dims))
        proto.dim.extend(
            []
        )  # Make sure that the field is set at least with an empty list.
        return proto

    @property
    def maybe_rank(self) -> Optional[int]:
        """Get the rank of this Shape, or None if it is unknown."""
        return len(self.dims) if self.dims is not None else None

    @property
    def rank(self) -> int:
        r = self.maybe_rank
        if r is None:
            raise ShapeError(f"Rank of {self} is unknown.")
        return r

    def __getitem__(self, item) -> Union["Shape", Natural]:
        """Indexing the dimensions, also provides iteration."""
        if self.dims is None:
            raise ShapeError(f"Cannot index unknown {self}.")
        elif isinstance(item, slice):
            return Shape(self.dims[item])
        return self.dims[item]

    def can_broadcast(self, other: "Shape") -> bool:
        """Check if this shape can be broadcast with ``other``."""
        try:
            self.broadcast(other)
        except ShapeError:
            return False
        else:
            return True

    def broadcast(self, other: Union["Shape", SimpleShape]) -> "Shape":
        """Return the result of shape broadcasting on ``self`` and ``other``."""
        if not isinstance(other, Shape):
            other = Shape.from_simple(other)
        a, b = self.to_simple(), other.to_simple()
        if a is None or b is None:
            return Shape(None)
        if len(a) > len(b):  # w.l.o.g. rank a < rank b
            a, b = b, a
        a = (1,) * (len(b) - len(a)) + a  # prepend with 1
        try:
            return Shape.from_simple(tuple(_broadcast_elem(x, y) for x, y in zip(a, b)))
        except ShapeError as e:
            raise ShapeError(
                f"Could not broadcast shapes: {self.to_simple()}, {other.to_simple()}."
            ) from e

    def __le__(self, other: "Shape") -> bool:
        """Shape membership comparison. Unknown shapes are treated as "any" qualifiers."""
        if not isinstance(other, Shape):
            return NotImplemented
        elif self.dims is None or other.dims is None:
            return True
        elif self.rank != other.rank:
            return False
        return all(x <= y for x, y in zip(self.dims, other.dims))


def _broadcast_elem(x: SimpleShapeElem, y: SimpleShapeElem) -> SimpleShapeElem:
    """
    Utility function for getting the broadcasting result over one dimension.
    Note that this is a very relaxed check, as unknowns may equal any constant or unknown (or they may be 1).
    Only constant (integer) dimensions are compared strictly.
    """
    if x == y:
        return x
    if x == 1:
        return y
    if y == 1:
        return x
    if isinstance(x, int) and isinstance(y, int):
        if x != y:
            raise ShapeError(
                f"Could not broadcast different constant dimensions: {x}, {y}."
            )
        return x
    if isinstance(x, int):
        return x
    if isinstance(y, int):
        return y
    return None



"""
Module for Spox operators that implement special internal behaviour that does not fit into the ONNX IR.
They behave like a normal Node, but their inference, building and translation behaviour may be overriden.
"""

from abc import ABC
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Set, Tuple

import onnx

from ._attributes import AttrString, AttrTensor, AttrType
from ._fields import BaseAttributes, BaseInputs, BaseOutputs
from ._node import Node, OpType
from ._scope import Scope
from ._shape import SimpleShape
from ._type_system import Tensor, Type
from ._value_prop import PropValueType
from ._var import Var

# This is a default used for internal operators that
# require the default domain. The most common of these
# is Introduce, which is effectively used in every graph.
# It implements an 'opset-independent' `Identity`.
#
# At the time of writing, there are three versions for
# `Identity` in the standard: 1, 13 and 14. 1 and 13 only
# cover `tensor`s. 14 includes `seq` as well. No version
# currently supports `map`. Another constraint at play is that
# ORT will throw warnings when encountering any version less
# than 7. If we want to avoid spamming the user with warnings
# in simple test cases, we ought to use 7 or higher.
#
# Version 7 was released some time in May 2018, version 13 in
# Nov 2020, and version 14 in April 2021. The only version
# that users can currently choose via `spox.opset.ai.onnx` is
# version 17 (released June 2022) and I doubt we will ever
# decrease that.
#
# Putting it all together, it seems reasonable to set this to
# 14. That version is almost 2 years old at this point and
# spares us any special casing if we want to use `seq`.
# This effectively sets the lower bound on every model.
INTERNAL_MIN_OPSET = 14


class _InternalNode(Node, ABC):
    @property
    def opset_req(self) -> Set[Tuple[str, int]]:
        return set()


class Argument(_InternalNode):
    """
    Internal operator representing the source of an argument.

    - The ``type`` has to be set, as otherwise the graph would be malformed.
    - If ``name`` is undeclared, it may be set to anything by the build (useful for subgraphs where order is used).
    - Additionally, an argument may have a ``default`` (an initializer) -
      but note that ONNX Runtime only supports non-overridable initializers (implemented by Initializer).
    """

    op_type = OpType("Argument", "spox.internal", 0)

    @dataclass
    class Attributes(BaseAttributes):
        type: AttrType
        name: Optional[AttrString] = None
        default: Optional[AttrTensor] = None

    @dataclass
    class Inputs(BaseInputs):
        pass

    @dataclass
    class Outputs(BaseOutputs):
        arg: Var

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs

    def post_init(self, **kwargs):
        if self.attrs.name is not None:
            self.outputs.arg._rename(self.attrs.name.value)

    def infer_output_types(self) -> Dict[str, Type]:
        # Output type is based on the value of the type attribute
        return {"arg": self.attrs.type.value}

    def update_metadata(self, opset_req, initializers, functions):
        super().update_metadata(opset_req, initializers, functions)
        var = self.outputs.arg
        if self.attrs.default is not None:
            initializers[var] = self.attrs.default.value

    def to_onnx(
        self, scope: "Scope", doc_string: Optional[str] = None, build_subgraph=None
    ) -> List[onnx.NodeProto]:
        return []


class _Initializer(_InternalNode):
    """Internal operator representing a non-overridable initializer."""

    op_type = OpType("Initializer", "spox.internal", 0)

    @dataclass
    class Attributes(BaseAttributes):
        value: AttrTensor

    @dataclass
    class Outputs(BaseOutputs):
        arg: Var

    attrs: Attributes
    inputs: BaseInputs
    outputs: Outputs

    def infer_output_types(self) -> Dict[str, Type]:
        # Output type is based on the value of the type attribute
        arr = self.attrs.value.value
        return {"arg": Tensor(arr.dtype, arr.shape)}

    def propagate_values(self) -> Dict[str, PropValueType]:
        return {"arg": self.attrs.value.value}

    def update_metadata(self, opset_req, initializers, functions):
        super().update_metadata(opset_req, initializers, functions)
        initializers[self.outputs.arg] = self.attrs.value.value

    def to_onnx(
        self, scope: "Scope", doc_string: Optional[str] = None, build_subgraph=None
    ) -> List[onnx.NodeProto]:
        # Initializers are added via update_metadata and don't affect the nodes proto list
        return []


class _Introduce(_InternalNode):
    """Internal operator used for introducing values, to manually evaluate them in the current scope."""

    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        inputs: Sequence[Var]

    @dataclass
    class Outputs(BaseOutputs):
        outputs: Sequence[Var]

    op_type = OpType("Introduce", "spox.internal", 0)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs

    def infer_output_types(self) -> Dict[str, Type]:
        return {
            f"outputs_{i}": arr.type
            for i, arr in enumerate(self.inputs.inputs)
            if arr.type is not None
        }

    @property
    def opset_req(self) -> Set[Tuple[str, int]]:
        return {("", INTERNAL_MIN_OPSET)}

    def to_onnx(
        self, scope: Scope, doc_string: Optional[str] = None, build_subgraph=None
    ) -> List[onnx.NodeProto]:
        assert len(self.inputs.inputs) == len(self.outputs.outputs)
        # Just create a renaming identity from what we forwarded into our actual output
        protos = []
        name = scope.node[self]
        for i in range(len(self.inputs.inputs)):
            protos.append(
                onnx.helper.make_node(
                    "Identity",
                    [scope.var[self.inputs.inputs[i]]],
                    [scope.var[self.outputs.outputs[i]]],
                    name + f"_id{i}",
                    doc_string,
                )
            )
        return protos


def intros(*args: Var) -> Sequence[Var]:
    """
    Internal identity operator with variadic arguments.

    As the underlying node is dependent on all passed arguments, this can be used to enforce specific evaluation order
    for values used in a subgraph - but otherwise ignored.

    For example, in a Loop whose body uses some ``x``, ``x`` may only be built within the subgraph and hence
    reevaluated on every iteration. If the Loop is wrapped with ``intro(x, loop(...))`` it is guaranteed that ``x``
    will be built outside of Loop's subgraph. It can be said that ``x`` was `introduced` in the outer scope.

    Parameters
    ----------
    args
        Vars to introduce in current scope.

    Returns
    -------
    Sequence[Var]
        Vars of the same value as ``args``, but with a shared dependency.
    """
    return _Introduce(
        None, _Introduce.Inputs(args), out_variadic=len(args)
    ).outputs.outputs


def intro(*args: Var) -> Var:
    """Introduces arguments like ``intros``, but only returns the last."""
    return intros(*args)[-1]


def unsafe_cast(x: Var, typ: Type) -> Var:
    """
    Creates a new var with the type forcefully set to ``typ``.

    Assumes that the real type of the Var is indeed compatible with ``shape`` (for example it was unknown).

    The function is meant for use when type inference failed, and it has to be overriden to avoid further failures.

    If you want to properly change a ``Var``'s type, use an operator like Cast, CastLike, Optional, etc.

    Parameters
    ----------
    x
        Var to retype.
    typ
        Target type - must be a constant.

    Returns
    -------
    Var
        Var with the type reset to whatever was given.
    """
    y = intro(x)
    y.type = typ
    y._value = x._value
    return y


def unsafe_reshape(x: Var, shape: SimpleShape) -> Var:
    """
    Creates a new var with the shape forcefully set to ``shape`` (like an unsafe cast).

    Assumes that the real shape of the Var is indeed compatible with ``shape`` (for example it was unknown).

    The function is meant for use when shape inference failed, and it has to be overriden to avoid failures.

    If you want to reshape to the shape of another var, use a Reshape operator.

    Parameters
    ----------
    x
        Var to reshape.
    shape
        Target shape - must be a constant.
    Returns
    -------
    Var
        Var with the same Tensor element type, but different shape.
    """
    return unsafe_cast(x, Tensor(x.unwrap_tensor().dtype, shape))



import numpy
import pytest

import spox.opset.ai.onnx.v17 as op
from spox._graph import arguments, initializer, results
from spox._internal_op import intro
from spox._type_system import Tensor


@pytest.fixture
def min_graph():
    first, second = arguments(
        first=Tensor(numpy.float32, (None,)), second=Tensor(numpy.float32, (None,))
    )
    return results(final=op.add(first, second))


@pytest.fixture
def square_graph():
    (value,) = arguments(value=Tensor(numpy.float32, (None,)))
    return results(final=op.identity(op.mul(value, value)))


@pytest.fixture
def trivial_seq_graph():
    first, second = arguments(
        first=Tensor(numpy.float32, (None,)), second=Tensor(numpy.float32, (None,))
    )
    return results(final=intro(op.add(first, second), second))


@pytest.fixture
def expanding_seq_graph():
    first, second = arguments(
        first=Tensor(numpy.float32, (None,)), second=Tensor(numpy.float32, (None,))
    )
    return results(final=intro(first, second, op.add(first, second)))


@pytest.fixture
def copy_graph():
    first, second = arguments(
        first=Tensor(numpy.float32, (None,)), second=Tensor(numpy.float64, (None,))
    )
    return results(final=first)


@pytest.fixture
def tri_graph():
    first, second, third = arguments(
        first=Tensor(numpy.float32, (2, None)),
        second=Tensor(numpy.float32, (None,)),
        third=Tensor(numpy.int32, (None,)),
    )
    return results(
        final=op.add(first, op.mul(second, op.cast(third, to=numpy.float32)))
    )


@pytest.fixture
def initializer_graph():
    normal, defaulted = arguments(
        normal=Tensor(numpy.float32, (None,)),
        defaulted=numpy.array([3, 2, 1], dtype=numpy.float32),
    )
    frozen = initializer(numpy.array([1, 10, 100], dtype=numpy.float32))
    return results(result=op.mul(op.add(normal, defaulted), frozen))


def test_minimal(onnx_helper, min_graph):
    onnx_helper.assert_close(
        onnx_helper.run(min_graph, "final", first=[1, 2, 3], second=[3, 2, 1]),
        [4, 4, 4],
    )
    onnx_helper.assert_close(
        onnx_helper.run(
            min_graph, "final", first=[1, 2, 3, 4, 5], second=[2, 4, 6, 8, 10]
        ),
        [3, 6, 9, 12, 15],
    )


def test_square(onnx_helper, square_graph):
    onnx_helper.assert_close(
        onnx_helper.run(square_graph, "final", value=[2, 3, 4]),
        [4, 9, 16],
    )


def test_trivial_seq(onnx_helper, trivial_seq_graph):
    # This will throw if ``first`` was removed from the graph due to ignored sequencing.
    onnx_helper.assert_close(
        onnx_helper.run(trivial_seq_graph, "final", first=[1, 2, 3], second=[3, 2, 1]),
        [3, 2, 1],
    )


def test_expanding_seq(onnx_helper, expanding_seq_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            expanding_seq_graph, "final", first=[1, 2, 3], second=[3, 2, 1]
        ),
        [4, 4, 4],
    )


def test_no_shape_throws():
    first, second = arguments(
        first=Tensor(numpy.float32, (None,)), second=Tensor(numpy.float32)
    )
    fun = results(final=op.add(first, second))
    with pytest.raises(ValueError):
        fun.to_onnx_model()


def test_no_type_throws():
    with pytest.raises(TypeError):
        first, second = arguments(first=Tensor(numpy.float32, (None,)), second=None)


def test_copy(onnx_helper, copy_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            copy_graph, "final", first=[1, 2, 3]
        ),  # Unused input is not in model
        [1, 2, 3],
    )


def test_triple_with_cast(onnx_helper, tri_graph):
    a = numpy.random.rand(2, 4).astype(numpy.float32)
    b = numpy.random.rand(4).astype(numpy.float32)
    c = numpy.array([1, 2, 3, 4], dtype=numpy.int32)
    onnx_helper.assert_close(
        onnx_helper.run(tri_graph, "final", first=a, second=b, third=c),
        a + b * c.astype(numpy.float32),
    )


def test_initializer(onnx_helper, initializer_graph):
    onnx_helper.assert_close(
        onnx_helper.run(initializer_graph, "result", normal=[1, 2, 3]),
        [4, 40, 400],
    )


@pytest.mark.xfail(
    reason="ONNX Runtime does not seem to support overriding initializers."
)
def test_initializer_override(onnx_helper, initializer_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            initializer_graph, "result", normal=[1, 2, 3], defaulted=[4, 3, 2]
        ),
        [5, 50, 500],
    )


def test_initializer_name_sets_correct(initializer_graph):
    proto = initializer_graph.to_onnx_model()
    assert {arg.name for arg in proto.graph.input} == {"normal", "defaulted"}
    assert "defaulted" in {arg.name for arg in proto.graph.initializer}



from typing import Callable

import numpy
import pytest

import spox.opset.ai.onnx.v17 as op
from spox._graph import arguments, results
from spox._type_system import Tensor
from spox._var import Var


@pytest.fixture(scope="session")
def optional_lifting_graph(ext):
    def liftA(f: Callable[[Var], Var], x: Var) -> Var:
        return ext.maybe(op.optional_has_element(x), f(op.optional_get_element(x)))

    def liftA2(f: Callable[[Var, Var], Var], x: Var, y: Var) -> Var:
        if x.type != y.type:
            raise TypeError(f"{x.type} != {y.type}")
        return ext.maybe(
            op.and_(op.optional_has_element(x), op.optional_has_element(y)),
            f(op.optional_get_element(x), op.optional_get_element(y)),
        )

    # Model: add the two arguments as float32, only if a is even, otherwise None.
    a, b = arguments(a=Tensor(numpy.int64, ()), b=Tensor(numpy.float32, ()))

    # Value only present if argument a is even
    is_even = op.reshape(
        op.not_(op.cast(op.mod(a, op.const(2)), to=numpy.bool_)), op.const([1])
    )
    a_even = ext.maybe(is_even, a)

    # Cast value to float32 (if it exists)
    a_even_f32 = liftA((lambda x: op.cast(x, to=numpy.float32)), a_even)

    # Add a_even cast to float32 and b (if a_even exists)
    r = liftA2(op.add, a_even_f32, op.optional(b))

    return results(r=r)


@pytest.mark.parametrize("a_value,b_value,r_value", [(8, 1.5, 9.5), (7, 1.5, None)])
def test_optional_lifting(
    onnx_helper, optional_lifting_graph, a_value, b_value, r_value
):
    onnx_helper.assert_close(
        onnx_helper.run(
            optional_lifting_graph,
            "r",
            a=numpy.array(a_value, dtype=numpy.int64),
            b=numpy.array(b_value, dtype=numpy.float32),
        ),
        r_value,
    )



import numpy as np
import pytest

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


@pytest.mark.parametrize(
    "keys_type,keys_kwarg,keys",
    [
        (np.float32, "keys_floats", (-0.5, 0.0, 0.5)),
        (np.int64, "keys_int64s", (1, 2, 3)),
        (np.str_, "keys_strings", ("x", "y", "z")),
    ],
)
@pytest.mark.parametrize(
    "values_type,values_kwarg,values",
    [
        (np.float32, "values_floats", (0.0, 0.5, 1.0)),
        (np.int64, "values_int64s", (0, 1, 2)),
        (np.str_, "values_strings", ("a", "b", "c")),
    ],
)
def test_label_encoder_inference(
    keys_type, keys_kwarg, keys, values_type, values_kwarg, values
):
    (x,) = arguments(x=Tensor(keys_type, (None, 5, "N")))
    y = op_ml.label_encoder(x, **{keys_kwarg: keys, values_kwarg: values})
    assert y.type == Tensor(values_type, (None, 5, "N"))



from copy import copy, deepcopy

import numpy as np
import onnx
import onnxruntime
import pytest

from spox import Tensor, argument, build, inline
from spox.opset.ai.onnx import v17 as op


@pytest.fixture
def simple_model():
    x, y = argument(Tensor(float, ())), argument(Tensor(float, ()))
    z = op.add(x, op.mul(y, op.cast(op.constant(value_float=2.0), to=float)))
    return build({"x": x, "y": y}, {"z": z})


def test_simple_build(simple_model):
    assert onnxruntime.InferenceSession(simple_model.SerializeToString()).run(
        None, {"x": np.array(7.0), "y": np.array(5.0)}
    ) == np.array(17.0)


def test_build_no_outputs_raises():
    with pytest.raises(Exception):
        build({"x": argument(Tensor(float, ()))}, {})


def test_build_inputs_not_vars_raises():
    x, y = argument(Tensor(float, ())), argument(Tensor(float, ()))
    with pytest.raises(TypeError):
        build({"x": x, "y": y}, {"z": None})  # type: ignore
    with pytest.raises(TypeError):
        build({"x": x, "y": y}, {"z": [op.add(x, y)]})  # type: ignore


def test_build_outputs_not_vars_raises():
    x, y = argument(Tensor(float, ())), argument(Tensor(float, ()))
    with pytest.raises(TypeError):
        build({"x": x, "y": y, "z": None}, {"z": op.add(x, y)})  # type: ignore
    with pytest.raises(TypeError):
        build({"x": x, "y": y, "xs": [x]}, {"z": op.add(x, y)})  # type: ignore


def test_build_inputs_not_arguments_raises():
    x, y = argument(Tensor(float, ())), argument(Tensor(float, ()))
    with pytest.raises(TypeError):
        build({"x": x, "y": op.add(x, y)}, {"z": op.add(x, y)})
    with pytest.raises(TypeError):
        build({"x": x, "y": y, "t": op.const(1)}, {"z": op.add(x, y)})


def test_simple_inline(simple_model):
    a, b = argument(Tensor(float, ())), argument(Tensor(float, ()))
    (c,) = inline(simple_model)(a, y=b).values()
    model = build({"a": a, "b": b}, {"c": c})
    assert onnxruntime.InferenceSession(model.SerializeToString()).run(
        None, {"a": np.array(7.0), "b": np.array(5.0)}
    ) == np.array(17.0)


def test_simple_inline_with_defaults(simple_model):
    model = onnx.ModelProto()
    model.ParseFromString(simple_model.SerializeToString())
    model.graph.initializer.append(
        onnx.helper.make_tensor("x", onnx.TensorProto.DOUBLE, (), [0.0])
    )
    b = argument(Tensor(float, ()))
    (c,) = inline(model)(y=b).values()
    model = build({"b": b}, {"c": c})
    assert onnxruntime.InferenceSession(model.SerializeToString()).run(
        None, {"b": np.array(5.0)}
    ) == np.array(10.0)


def test_simple_inline_types(simple_model):
    a = argument(Tensor(float, ()))
    (r,) = inline(simple_model)(a, a).values()
    assert r.type == Tensor(float, ())


def test_simple_inline_bad_types(simple_model):
    a = argument(Tensor(float, ()))
    b = argument(Tensor(float, ("N", 3)))
    c = argument(Tensor(int, ()))
    with pytest.raises(TypeError):
        inline(simple_model)(b, a)
    with pytest.raises(TypeError):
        inline(simple_model)(a, c)


def test_shallow_copy_var():
    a = argument(Tensor(float, ()))
    b = op.add(a, copy(a))
    build({"a": a}, {"b": b})


def test_shallow_deepcopy_var_raises():
    a = argument(Tensor(float, ()))
    with pytest.raises(ValueError):
        deepcopy(a)



# ruff: noqa: E741 -- Allow ambiguous variable name
from dataclasses import dataclass
from typing import (
    Dict,
    Iterable,
    Optional,
    Sequence,
    Tuple,
)

import numpy as np

from spox._attributes import (
    AttrFloat32,
    AttrFloat32s,
    AttrInt64,
    AttrInt64s,
    AttrString,
    AttrStrings,
    AttrTensor,
)
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._node import OpType
from spox._standard import InferenceError, StandardNode
from spox._type_system import Tensor, Type
from spox._var import Var


class _ArrayFeatureExtractor(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        X: Var
        Y: Var

    @dataclass
    class Outputs(BaseOutputs):
        Z: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if not self.inputs.fully_typed:
            return {}
        xt, yt = self.inputs.X.unwrap_tensor(), self.inputs.Y.unwrap_tensor()
        assert xt.shape is not None  # already checked with fully_typed
        assert yt.shape is not None  # already checked with fully_typed
        if len(xt.shape) < 1:
            raise InferenceError("Expected rank >= 1")
        if len(yt.shape) != 1:
            raise InferenceError("Input `Y` must be of rank 1.")
        if len(xt.shape) == 1:
            return {"Z": Tensor(xt.dtype, (1, yt.shape[-1]))}
        shape = tuple(list(xt.shape[:-1]) + [yt.shape[-1]])  # type: ignore
        return {"Z": Tensor(xt.dtype, shape)}

    op_type = OpType("ArrayFeatureExtractor", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _Binarizer(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        threshold: AttrFloat32

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        return {"Y": self.inputs.X.type} if self.inputs.X.type is not None else {}

    op_type = OpType("Binarizer", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _CastMap(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        cast_to: AttrString
        map_form: AttrString
        max_map: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    op_type = OpType("CastMap", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _CategoryMapper(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        cats_int64s: Optional[AttrInt64s]
        cats_strings: Optional[AttrStrings]
        default_int64: AttrInt64
        default_string: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if not self.inputs.fully_typed:
            return {}
        cats1, cats2 = self.attrs.cats_int64s, self.attrs.cats_strings
        if cats1 is None or cats2 is None:
            raise InferenceError("Missing required attributes.")
        if len(cats1.value) != len(cats2.value):
            raise InferenceError("Categories lists have mismatched lengths.")
        t = self.inputs.X.unwrap_tensor()
        (elem_type,) = {np.int64, np.str_} - {t.dtype.type}  # type: ignore
        return {"Y": Tensor(elem_type, t.shape)}

    op_type = OpType("CategoryMapper", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _DictVectorizer(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        int64_vocabulary: Optional[AttrInt64s]
        string_vocabulary: Optional[AttrStrings]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    op_type = OpType("DictVectorizer", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _FeatureVectorizer(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        inputdimensions: Optional[AttrInt64s]

    @dataclass
    class Inputs(BaseInputs):
        X: Sequence[Var]

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    op_type = OpType("FeatureVectorizer", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _Imputer(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        imputed_value_floats: Optional[AttrFloat32s]
        imputed_value_int64s: Optional[AttrInt64s]
        replaced_value_float: AttrFloat32
        replaced_value_int64: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if not self.inputs.fully_typed:
            return {}
        t = self.inputs.X.unwrap_tensor()
        # We verify if the attributes are set correctly and matching the input elem type
        cases = {
            np.int64: (
                self.attrs.imputed_value_int64s,
                self.attrs.replaced_value_int64,
            ),
            np.float32: (
                self.attrs.imputed_value_floats,
                self.attrs.replaced_value_float,
            ),
        }
        for key, (imp, rep) in cases.items():
            if t.dtype.type is key:
                if not all(
                    imp1 is None for key1, (imp1, rep1) in cases.items() if key != key1
                ):
                    raise InferenceError("Only one input imputed type may be set.")
                break
        else:
            raise InferenceError("No matching element type")
        if imp is None:
            raise InferenceError("Value list for imputation is required.")
        # If the number of features is known (last row, we can check this here)
        sim = t.shape
        last = sim[-1] if sim else 1
        if isinstance(last, int) and len(imp.value) not in {1, last}:
            raise InferenceError(
                f"Mismatched expected ({len(imp.value)}) and actual ({last}) feature count."
            )
        return {"Y": t}

    op_type = OpType("Imputer", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _LabelEncoder(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        default_float: AttrFloat32
        default_int64: AttrInt64
        default_string: AttrString
        keys_floats: Optional[AttrFloat32s]
        keys_int64s: Optional[AttrInt64s]
        keys_strings: Optional[AttrStrings]
        values_floats: Optional[AttrFloat32s]
        values_int64s: Optional[AttrInt64s]
        values_strings: Optional[AttrStrings]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    op_type = OpType("LabelEncoder", "ai.onnx.ml", 2)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _LinearClassifier(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        classlabels_ints: Optional[AttrInt64s]
        classlabels_strings: Optional[AttrStrings]
        coefficients: AttrFloat32s
        intercepts: Optional[AttrFloat32s]
        multi_class: AttrInt64
        post_transform: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var
        Z: Var

    op_type = OpType("LinearClassifier", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _LinearRegressor(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        coefficients: Optional[AttrFloat32s]
        intercepts: Optional[AttrFloat32s]
        post_transform: AttrString
        targets: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if not self.inputs.fully_typed:
            return {}
        sim = self.inputs.X.unwrap_tensor().shape
        assert sim is not None
        if len(sim) == 2:
            return {"Y": Tensor(np.float32, sim)}
        elif len(sim) == 1:
            return {"Y": Tensor(np.float32, (1, sim[0]))}
        elif len(sim) == 0:
            return {"Y": Tensor(np.float32, (1, 1))}
        else:
            raise InferenceError("Input shape must be at most a matrix.")

    op_type = OpType("LinearRegressor", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _Normalizer(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        norm: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if self.attrs.norm.value not in ("MAX", "L1", "L2"):
            raise InferenceError(
                f"Unknown normalisation method `{self.attrs.norm.value}`"
            )
        return {"Y": self.inputs.X.type} if self.inputs.X.type is not None else {}

    op_type = OpType("Normalizer", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _OneHotEncoder(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        cats_int64s: Optional[AttrInt64s]
        cats_strings: Optional[AttrStrings]
        zeros: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if not self.inputs.fully_typed:
            return {}
        if self.attrs.cats_int64s:
            n_encodings = len(self.attrs.cats_int64s.value)
        elif self.attrs.cats_strings:
            n_encodings = len(self.attrs.cats_strings.value)
        else:
            raise InferenceError(
                "Either `cats_int64s` or `cats_strings` attributes must be set."
            )
        shape = (*self.inputs.X.unwrap_tensor().shape, n_encodings)  # type: ignore
        return {"Y": Tensor(dtype=np.float32, shape=shape)}

    op_type = OpType("OneHotEncoder", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _SVMClassifier(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        classlabels_ints: Optional[AttrInt64s]
        classlabels_strings: Optional[AttrStrings]
        coefficients: Optional[AttrFloat32s]
        kernel_params: Optional[AttrFloat32s]
        kernel_type: AttrString
        post_transform: AttrString
        prob_a: Optional[AttrFloat32s]
        prob_b: Optional[AttrFloat32s]
        rho: Optional[AttrFloat32s]
        support_vectors: Optional[AttrFloat32s]
        vectors_per_class: Optional[AttrInt64s]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var
        Z: Var

    op_type = OpType("SVMClassifier", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _SVMRegressor(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        coefficients: Optional[AttrFloat32s]
        kernel_params: Optional[AttrFloat32s]
        kernel_type: AttrString
        n_supports: AttrInt64
        one_class: AttrInt64
        post_transform: AttrString
        rho: Optional[AttrFloat32s]
        support_vectors: Optional[AttrFloat32s]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    op_type = OpType("SVMRegressor", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _Scaler(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        offset: Optional[AttrFloat32s]
        scale: Optional[AttrFloat32s]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if self.inputs.X.type is None:
            return {}
        sc, off = self.attrs.scale, self.attrs.offset
        if sc is None or off is None:
            raise InferenceError("Scale and offset are required attributes.")
        t = self.inputs.X.unwrap_tensor()
        # If the number of features is known (last row, we can check this here)
        last = t.shape[-1] if t.shape else 1
        if isinstance(last, int) and len(sc.value) not in {1, last}:
            raise InferenceError(
                f"Mismatched expected ({len(sc.value)}) and actual ({last}) feature count for scale."
            )
        if isinstance(last, int) and len(off.value) not in {1, last}:
            raise InferenceError(
                f"Mismatched expected ({len(off.value)}) and actual ({last}) feature count for offset."
            )
        return {"Y": Tensor(np.float32, t.shape)}

    op_type = OpType("Scaler", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _TreeEnsembleClassifier(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        base_values: Optional[AttrFloat32s]
        base_values_as_tensor: Optional[AttrTensor]
        class_ids: Optional[AttrInt64s]
        class_nodeids: Optional[AttrInt64s]
        class_treeids: Optional[AttrInt64s]
        class_weights: Optional[AttrFloat32s]
        class_weights_as_tensor: Optional[AttrTensor]
        classlabels_int64s: Optional[AttrInt64s]
        classlabels_strings: Optional[AttrStrings]
        nodes_falsenodeids: Optional[AttrInt64s]
        nodes_featureids: Optional[AttrInt64s]
        nodes_hitrates: Optional[AttrFloat32s]
        nodes_hitrates_as_tensor: Optional[AttrTensor]
        nodes_missing_value_tracks_true: Optional[AttrInt64s]
        nodes_modes: Optional[AttrStrings]
        nodes_nodeids: Optional[AttrInt64s]
        nodes_treeids: Optional[AttrInt64s]
        nodes_truenodeids: Optional[AttrInt64s]
        nodes_values: Optional[AttrFloat32s]
        nodes_values_as_tensor: Optional[AttrTensor]
        post_transform: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var
        Z: Var

    def infer_output_types(self) -> Dict[str, Type]:
        e = (
            len(self.attrs.class_ids.value)
            if self.attrs.class_ids is not None
            else None
        )
        if self.attrs.classlabels_strings is not None:
            y_type = np.str_
        elif self.attrs.classlabels_int64s is not None:
            y_type = np.int64  # type: ignore
        else:
            raise InferenceError(
                "Either string or int64 class labels should be defined"
            )
        if self.inputs.fully_typed:
            shape = self.inputs.X.unwrap_tensor().shape
            assert shape is not None  # already checked with fully_typed
            if len(shape) != 2:
                raise InferenceError("Expected input to be a matrix.")
            n = shape[0]
        else:
            n = None
        return {"Y": Tensor(y_type, (n,)), "Z": Tensor(np.float32, (n, e))}

    op_type = OpType("TreeEnsembleClassifier", "ai.onnx.ml", 3)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _TreeEnsembleRegressor(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        aggregate_function: AttrString
        base_values: Optional[AttrFloat32s]
        base_values_as_tensor: Optional[AttrTensor]
        n_targets: Optional[AttrInt64]
        nodes_falsenodeids: Optional[AttrInt64s]
        nodes_featureids: Optional[AttrInt64s]
        nodes_hitrates: Optional[AttrFloat32s]
        nodes_hitrates_as_tensor: Optional[AttrTensor]
        nodes_missing_value_tracks_true: Optional[AttrInt64s]
        nodes_modes: Optional[AttrStrings]
        nodes_nodeids: Optional[AttrInt64s]
        nodes_treeids: Optional[AttrInt64s]
        nodes_truenodeids: Optional[AttrInt64s]
        nodes_values: Optional[AttrFloat32s]
        nodes_values_as_tensor: Optional[AttrTensor]
        post_transform: AttrString
        target_ids: Optional[AttrInt64s]
        target_nodeids: Optional[AttrInt64s]
        target_treeids: Optional[AttrInt64s]
        target_weights: Optional[AttrFloat32s]
        target_weights_as_tensor: Optional[AttrTensor]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Y: Var

    def infer_output_types(self) -> Dict[str, Type]:
        if self.inputs.fully_typed:
            shape = self.inputs.X.unwrap_tensor().shape
            assert shape is not None  # already checked with fully_typed
            if len(shape) != 2:
                raise InferenceError("Expected input to be a matrix.")
            assert shape is not None
            n = shape[0]
        else:
            n = None
        e = self.attrs.n_targets.value if self.attrs.n_targets is not None else None
        return {"Y": Tensor(np.float32, (n, e))}

    op_type = OpType("TreeEnsembleRegressor", "ai.onnx.ml", 3)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _ZipMap(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        classlabels_int64s: Optional[AttrInt64s]
        classlabels_strings: Optional[AttrStrings]

    @dataclass
    class Inputs(BaseInputs):
        X: Var

    @dataclass
    class Outputs(BaseOutputs):
        Z: Var

    op_type = OpType("ZipMap", "ai.onnx.ml", 1)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


def array_feature_extractor(
    X: Var,
    Y: Var,
) -> Var:
    r"""
    Select elements of the input tensor based on the indices passed. The
    indices are applied to the last axes of the tensor.

    Parameters
    ==========
    X
        Type T.
        Data to be selected
    Y
        Type tensor(int64).
        The indices, based on 0 as the first index of any dimension.

    Returns
    =======
    Z : Var
        Type T.
        Selected output data as an array

    Notes
    =====
    Signature: ``ai.onnx.ml@1::ArrayFeatureExtractor``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`, `tensor(string)`
    """
    return _ArrayFeatureExtractor(
        _ArrayFeatureExtractor.Attributes(),
        _ArrayFeatureExtractor.Inputs(
            X=X,
            Y=Y,
        ),
    ).outputs.Z


def binarizer(
    X: Var,
    *,
    threshold: float = 0.0,
) -> Var:
    r"""
    Maps the values of the input tensor to either 0 or 1, element-wise,
    based on the outcome of a comparison against a threshold value.

    Parameters
    ==========
    X
        Type T.
        Data to be binarized
    threshold
        Attribute.
        Values greater than this are mapped to 1, others to 0.

    Returns
    =======
    Y : Var
        Type T.
        Binarized output data

    Notes
    =====
    Signature: ``ai.onnx.ml@1::Binarizer``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _Binarizer(
        _Binarizer.Attributes(
            threshold=AttrFloat32(threshold, name="threshold"),
        ),
        _Binarizer.Inputs(
            X=X,
        ),
    ).outputs.Y


def cast_map(
    X: Var,
    *,
    cast_to: str = "TO_FLOAT",
    map_form: str = "DENSE",
    max_map: int = 1,
) -> Var:
    r"""
    Converts a map to a tensor.The map key must be an int64 and the values
    will be ordered in ascending order based on this key.The operator
    supports dense packing or sparse packing. If using sparse packing, the
    key cannot exceed the max_map-1 value.

    Parameters
    ==========
    X
        Type T1.
        The input map that is to be cast to a tensor
    cast_to
        Attribute.
        A string indicating the desired element type of the output tensor, one
        of 'TO_FLOAT', 'TO_STRING', 'TO_INT64'.
    map_form
        Attribute.
        Indicates whether to only output as many values as are in the input
        (dense), or position the input based on using the key of the map as the
        index of the output (sparse).One of 'DENSE', 'SPARSE'.
    max_map
        Attribute.
        If the value of map_form is 'SPARSE,' this attribute indicates the total
        length of the output tensor.

    Returns
    =======
    Y : Var
        Type T2.
        A tensor representing the same data as the input map, ordered by their
        keys

    Notes
    =====
    Signature: ``ai.onnx.ml@1::CastMap``.

    Type constraints:
     - T1: `map(int64,tensor(float))`, `map(int64,tensor(string))`
     - T2: `tensor(float)`, `tensor(int64)`, `tensor(string)`
    """
    return _CastMap(
        _CastMap.Attributes(
            cast_to=AttrString(cast_to, name="cast_to"),
            map_form=AttrString(map_form, name="map_form"),
            max_map=AttrInt64(max_map, name="max_map"),
        ),
        _CastMap.Inputs(
            X=X,
        ),
    ).outputs.Y


def category_mapper(
    X: Var,
    *,
    cats_int64s: Optional[Iterable[int]] = None,
    cats_strings: Optional[Iterable[str]] = None,
    default_int64: int = -1,
    default_string: str = "_Unused",
) -> Var:
    r"""
    Converts strings to integers and vice versa. Two sequences of equal
    length are used to map between integers and strings, with strings and
    integers at the same index detailing the mapping. Each operator converts
    either integers to strings or strings to integers, depending on which
    default value attribute is provided. Only one default value attribute
    should be defined. If the string default value is set, it will convert
    integers to strings. If the int default value is set, it will convert
    strings to integers.

    Parameters
    ==========
    X
        Type T1.
        Input data
    cats_int64s
        Attribute.
        The integers of the map. This sequence must be the same length as the
        'cats_strings' sequence.
    cats_strings
        Attribute.
        The strings of the map. This sequence must be the same length as the
        'cats_int64s' sequence
    default_int64
        Attribute.
        An integer to use when an input string value is not found in the map.One
        and only one of the 'default\_\*' attributes must be defined.
    default_string
        Attribute.
        A string to use when an input integer value is not found in the map.One
        and only one of the 'default\_\*' attributes must be defined.

    Returns
    =======
    Y : Var
        Type T2.
        Output data. If strings are input, the output values are integers, and
        vice versa.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::CategoryMapper``.

    Type constraints:
     - T1: `tensor(int64)`, `tensor(string)`
     - T2: `tensor(int64)`, `tensor(string)`
    """
    return _CategoryMapper(
        _CategoryMapper.Attributes(
            cats_int64s=AttrInt64s.maybe(cats_int64s, name="cats_int64s"),
            cats_strings=AttrStrings.maybe(cats_strings, name="cats_strings"),
            default_int64=AttrInt64(default_int64, name="default_int64"),
            default_string=AttrString(default_string, name="default_string"),
        ),
        _CategoryMapper.Inputs(
            X=X,
        ),
    ).outputs.Y


def dict_vectorizer(
    X: Var,
    *,
    int64_vocabulary: Optional[Iterable[int]] = None,
    string_vocabulary: Optional[Iterable[str]] = None,
) -> Var:
    r"""
    Uses an index mapping to convert a dictionary to an array. Given a
    dictionary, each key is looked up in the vocabulary attribute
    corresponding to the key type. The index into the vocabulary array at
    which the key is found is then used to index the output 1-D tensor 'Y'
    and insert into it the value found in the dictionary 'X'. The key type
    of the input map must correspond to the element type of the defined
    vocabulary attribute. Therefore, the output array will be equal in
    length to the index mapping vector parameter. All keys in the input
    dictionary must be present in the index mapping vector. For each item in
    the input dictionary, insert its value in the output array. Any keys not
    present in the input dictionary, will be zero in the output array. For
    example: if the ``string_vocabulary`` parameter is set to
    ``["a", "c", "b", "z"]``, then an input of ``{"a": 4, "c": 8}`` will
    produce an output of ``[4, 8, 0, 0]``.

    Parameters
    ==========
    X
        Type T1.
        A dictionary.
    int64_vocabulary
        Attribute.
        An integer vocabulary array.One and only one of the vocabularies must be
        defined.
    string_vocabulary
        Attribute.
        A string vocabulary array.One and only one of the vocabularies must be
        defined.

    Returns
    =======
    Y : Var
        Type T2.
        A 1-D tensor holding values from the input dictionary.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::DictVectorizer``.

    Type constraints:
     - T1: `map(int64,tensor(double))`, `map(int64,tensor(float))`, `map(int64,tensor(string))`, `map(string,tensor(double))`, `map(string,tensor(float))`, `map(string,tensor(int64))`
     - T2: `tensor(double)`, `tensor(float)`, `tensor(int64)`, `tensor(string)`
    """
    return _DictVectorizer(
        _DictVectorizer.Attributes(
            int64_vocabulary=AttrInt64s.maybe(
                int64_vocabulary, name="int64_vocabulary"
            ),
            string_vocabulary=AttrStrings.maybe(
                string_vocabulary, name="string_vocabulary"
            ),
        ),
        _DictVectorizer.Inputs(
            X=X,
        ),
    ).outputs.Y


def feature_vectorizer(
    X: Sequence[Var],
    *,
    inputdimensions: Optional[Iterable[int]] = None,
) -> Var:
    r"""
    Concatenates input tensors into one continuous output. All input shapes
    are 2-D and are concatenated along the second dimension. 1-D tensors are
    treated as [1,C]. Inputs are copied to the output maintaining the order
    of the input arguments. All inputs must be integers or floats, while the
    output will be all floating point values.

    Parameters
    ==========
    X
        Type T1.
        An ordered collection of tensors, all with the same element type.
    inputdimensions
        Attribute.
        The size of each input in the input list

    Returns
    =======
    Y : Var
        Type tensor(float).
        The output array, elements ordered as the inputs.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::FeatureVectorizer``.

    Type constraints:
     - T1: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _FeatureVectorizer(
        _FeatureVectorizer.Attributes(
            inputdimensions=AttrInt64s.maybe(inputdimensions, name="inputdimensions"),
        ),
        _FeatureVectorizer.Inputs(
            X=X,
        ),
    ).outputs.Y


def imputer(
    X: Var,
    *,
    imputed_value_floats: Optional[Iterable[float]] = None,
    imputed_value_int64s: Optional[Iterable[int]] = None,
    replaced_value_float: float = 0.0,
    replaced_value_int64: int = 0,
) -> Var:
    r"""
    Replaces inputs that equal one value with another, leaving all other
    elements alone. This operator is typically used to replace missing
    values in situations where they have a canonical representation, such as
    -1, 0, NaN, or some extreme value. One and only one of
    imputed_value_floats or imputed_value_int64s should be defined -- floats
    if the input tensor holds floats, integers if the input tensor holds
    integers. The imputed values must all fit within the width of the tensor
    element type. One and only one of the replaced_value_float or
    replaced_value_int64 should be defined, which one depends on whether
    floats or integers are being processed. The imputed_value attribute
    length can be 1 element, or it can have one element per input feature.In
    other words, if the input tensor has the shape [\*,F], then the length
    of the attribute array may be 1 or F. If it is 1, then it is broadcast
    along the last dimension and applied to each feature.

    Parameters
    ==========
    X
        Type T.
        Data to be processed.
    imputed_value_floats
        Attribute.
        Value(s) to change to
    imputed_value_int64s
        Attribute.
        Value(s) to change to.
    replaced_value_float
        Attribute.
        A value that needs replacing.
    replaced_value_int64
        Attribute.
        A value that needs replacing.

    Returns
    =======
    Y : Var
        Type T.
        Imputed output data

    Notes
    =====
    Signature: ``ai.onnx.ml@1::Imputer``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _Imputer(
        _Imputer.Attributes(
            imputed_value_floats=AttrFloat32s.maybe(
                imputed_value_floats, name="imputed_value_floats"
            ),
            imputed_value_int64s=AttrInt64s.maybe(
                imputed_value_int64s, name="imputed_value_int64s"
            ),
            replaced_value_float=AttrFloat32(
                replaced_value_float, name="replaced_value_float"
            ),
            replaced_value_int64=AttrInt64(
                replaced_value_int64, name="replaced_value_int64"
            ),
        ),
        _Imputer.Inputs(
            X=X,
        ),
    ).outputs.Y


def label_encoder(
    X: Var,
    *,
    default_float: float = -0.0,
    default_int64: int = -1,
    default_string: str = "_Unused",
    keys_floats: Optional[Iterable[float]] = None,
    keys_int64s: Optional[Iterable[int]] = None,
    keys_strings: Optional[Iterable[str]] = None,
    values_floats: Optional[Iterable[float]] = None,
    values_int64s: Optional[Iterable[int]] = None,
    values_strings: Optional[Iterable[str]] = None,
) -> Var:
    r"""
    Maps each element in the input tensor to another value. The mapping is
    determined by the two parallel attributes, 'keys\_\ *' and 'values\_*'
    attribute. The i-th value in the specified 'keys\_\ *' attribute would
    be mapped to the i-th value in the specified 'values\_*' attribute. It
    implies that input's element type and the element type of the specified
    'keys\_\ *' should be identical while the output type is identical to
    the specified 'values\_*' attribute. If an input element can not be
    found in the specified 'keys\_\ *' attribute, the 'default\_*' that
    matches the specified 'values\_\ *' attribute may be used as its output
    value. Let's consider an example which maps a string tensor to an
    integer tensor. Assume and 'keys_strings' is ["Amy", "Sally"],
    'values_int64s' is [5, 6], and 'default_int64' is '-1'. The input
    ["Dori", "Amy", "Amy", "Sally", "Sally"] would be mapped to [-1, 5, 5,
    6, 6]. Since this operator is an one-to-one mapping, its input and
    output shapes are the same. Notice that only one of
    'keys\_*'/'values\_\ *' can be set. For key look-up, bit-wise comparison
    is used so even a float NaN can be mapped to a value in 'values\_*'
    attribute.

    Parameters
    ==========
    X
        Type T1.
        Input data. It can be either tensor or scalar.
    default_float
        Attribute.
        A float.
    default_int64
        Attribute.
        An integer.
    default_string
        Attribute.
        A string.
    keys_floats
        Attribute.
        A list of floats.
    keys_int64s
        Attribute.
        A list of ints.
    keys_strings
        Attribute.
        A list of strings. One and only one of 'keys\_\*'s should be set.
    values_floats
        Attribute.
        A list of floats.
    values_int64s
        Attribute.
        A list of ints.
    values_strings
        Attribute.
        A list of strings. One and only one of 'value\_\*'s should be set.

    Returns
    =======
    Y : Var
        Type T2.
        Output data.

    Notes
    =====
    Signature: ``ai.onnx.ml@2::LabelEncoder``.

    Type constraints:
     - T1: `tensor(float)`, `tensor(int64)`, `tensor(string)`
     - T2: `tensor(float)`, `tensor(int64)`, `tensor(string)`
    """
    return _LabelEncoder(
        _LabelEncoder.Attributes(
            default_float=AttrFloat32(default_float, name="default_float"),
            default_int64=AttrInt64(default_int64, name="default_int64"),
            default_string=AttrString(default_string, name="default_string"),
            keys_floats=AttrFloat32s.maybe(keys_floats, name="keys_floats"),
            keys_int64s=AttrInt64s.maybe(keys_int64s, name="keys_int64s"),
            keys_strings=AttrStrings.maybe(keys_strings, name="keys_strings"),
            values_floats=AttrFloat32s.maybe(values_floats, name="values_floats"),
            values_int64s=AttrInt64s.maybe(values_int64s, name="values_int64s"),
            values_strings=AttrStrings.maybe(values_strings, name="values_strings"),
        ),
        _LabelEncoder.Inputs(
            X=X,
        ),
    ).outputs.Y


def linear_classifier(
    X: Var,
    *,
    classlabels_ints: Optional[Iterable[int]] = None,
    classlabels_strings: Optional[Iterable[str]] = None,
    coefficients: Iterable[float],
    intercepts: Optional[Iterable[float]] = None,
    multi_class: int = 0,
    post_transform: str = "NONE",
) -> Tuple[Var, Var]:
    r"""
    Linear classifier

    Parameters
    ==========
    X
        Type T1.
        Data to be classified.
    classlabels_ints
        Attribute.
        Class labels when using integer labels. One and only one 'classlabels'
        attribute must be defined.
    classlabels_strings
        Attribute.
        Class labels when using string labels. One and only one 'classlabels'
        attribute must be defined.
    coefficients
        Attribute.
        A collection of weights of the model(s).
    intercepts
        Attribute.
        A collection of intercepts.
    multi_class
        Attribute.
        Indicates whether to do OvR or multinomial (0=OvR is the default).
    post_transform
        Attribute.
        Indicates the transform to apply to the scores vector.One of 'NONE,'
        'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'

    Returns
    =======
    Y : Var
        Type T2.
        Classification outputs (one class per example).
    Z : Var
        Type tensor(float).
        Classification scores ([N,E] - one score for each class and example

    Notes
    =====
    Signature: ``ai.onnx.ml@1::LinearClassifier``.

    Type constraints:
     - T1: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
     - T2: `tensor(int64)`, `tensor(string)`
    """
    return _LinearClassifier(
        _LinearClassifier.Attributes(
            classlabels_ints=AttrInt64s.maybe(
                classlabels_ints, name="classlabels_ints"
            ),
            classlabels_strings=AttrStrings.maybe(
                classlabels_strings, name="classlabels_strings"
            ),
            coefficients=AttrFloat32s(coefficients, name="coefficients"),
            intercepts=AttrFloat32s.maybe(intercepts, name="intercepts"),
            multi_class=AttrInt64(multi_class, name="multi_class"),
            post_transform=AttrString(post_transform, name="post_transform"),
        ),
        _LinearClassifier.Inputs(
            X=X,
        ),
    ).outputs._unpack_to_any()


def linear_regressor(
    X: Var,
    *,
    coefficients: Optional[Iterable[float]] = None,
    intercepts: Optional[Iterable[float]] = None,
    post_transform: str = "NONE",
    targets: int = 1,
) -> Var:
    r"""
    Generalized linear regression evaluation. If targets is set to 1
    (default) then univariate regression is performed. If targets is set to
    M then M sets of coefficients must be passed in as a sequence and M
    results will be output for each input n in N. The coefficients array is
    of length n, and the coefficients for each target are contiguous.
    Intercepts are optional but if provided must match the number of
    targets.

    Parameters
    ==========
    X
        Type T.
        Data to be regressed.
    coefficients
        Attribute.
        Weights of the model(s).
    intercepts
        Attribute.
        Weights of the intercepts, if used.
    post_transform
        Attribute.
        Indicates the transform to apply to the regression output vector.One of
        'NONE,' 'SOFTMAX,' 'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'
    targets
        Attribute.
        The total number of regression targets, 1 if not defined.

    Returns
    =======
    Y : Var
        Type tensor(float).
        Regression outputs (one per target, per example).

    Notes
    =====
    Signature: ``ai.onnx.ml@1::LinearRegressor``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _LinearRegressor(
        _LinearRegressor.Attributes(
            coefficients=AttrFloat32s.maybe(coefficients, name="coefficients"),
            intercepts=AttrFloat32s.maybe(intercepts, name="intercepts"),
            post_transform=AttrString(post_transform, name="post_transform"),
            targets=AttrInt64(targets, name="targets"),
        ),
        _LinearRegressor.Inputs(
            X=X,
        ),
    ).outputs.Y


def normalizer(
    X: Var,
    *,
    norm: str = "MAX",
) -> Var:
    r"""
    Normalize the input. There are three normalization modes, which have the
    corresponding formulas, defined using element-wise infix operators '/'
    and '^' and tensor-wide functions 'max' and 'sum': Max: Y = X / max(X)
    L1: Y = X / sum(X) L2: Y = sqrt(X^2 / sum(X^2)} In all modes, if the
    divisor is zero, Y == X. For batches, that is, [N,C] tensors,
    normalization is done along the C axis. In other words, each row of the
    batch is normalized independently.

    Parameters
    ==========
    X
        Type T.
        Data to be encoded, a tensor of shape [N,C] or [C]
    norm
        Attribute.
        One of 'MAX,' 'L1,' 'L2'

    Returns
    =======
    Y : Var
        Type tensor(float).
        Encoded output data

    Notes
    =====
    Signature: ``ai.onnx.ml@1::Normalizer``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _Normalizer(
        _Normalizer.Attributes(
            norm=AttrString(norm, name="norm"),
        ),
        _Normalizer.Inputs(
            X=X,
        ),
    ).outputs.Y


def one_hot_encoder(
    X: Var,
    *,
    cats_int64s: Optional[Iterable[int]] = None,
    cats_strings: Optional[Iterable[str]] = None,
    zeros: int = 1,
) -> Var:
    r"""
    Replace each input element with an array of ones and zeros, where a
    single one is placed at the index of the category that was passed in.
    The total category count will determine the size of the extra dimension
    of the output array Y. For example, if we pass a tensor with a single
    value of 4, and a category count of 8, the output will be a tensor with
    ``[0,0,0,0,1,0,0,0]``. This operator assumes every input feature is from
    the same set of categories. If the input is a tensor of float, int32, or
    double, the data will be cast to integers and the cats_int64s category
    list will be used for the lookups.

    Parameters
    ==========
    X
        Type T.
        Data to be encoded.
    cats_int64s
        Attribute.
        List of categories, ints.One and only one of the 'cats\_\*' attributes
        must be defined.
    cats_strings
        Attribute.
        List of categories, strings.One and only one of the 'cats\_\*'
        attributes must be defined.
    zeros
        Attribute.
        If true and category is not present, will return all zeros; if false and
        a category if not found, the operator will fail.

    Returns
    =======
    Y : Var
        Type tensor(float).
        Encoded output data, having one more dimension than X.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::OneHotEncoder``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`, `tensor(string)`
    """
    return _OneHotEncoder(
        _OneHotEncoder.Attributes(
            cats_int64s=AttrInt64s.maybe(cats_int64s, name="cats_int64s"),
            cats_strings=AttrStrings.maybe(cats_strings, name="cats_strings"),
            zeros=AttrInt64(zeros, name="zeros"),
        ),
        _OneHotEncoder.Inputs(
            X=X,
        ),
    ).outputs.Y


def svmclassifier(
    X: Var,
    *,
    classlabels_ints: Optional[Iterable[int]] = None,
    classlabels_strings: Optional[Iterable[str]] = None,
    coefficients: Optional[Iterable[float]] = None,
    kernel_params: Optional[Iterable[float]] = None,
    kernel_type: str = "LINEAR",
    post_transform: str = "NONE",
    prob_a: Optional[Iterable[float]] = None,
    prob_b: Optional[Iterable[float]] = None,
    rho: Optional[Iterable[float]] = None,
    support_vectors: Optional[Iterable[float]] = None,
    vectors_per_class: Optional[Iterable[int]] = None,
) -> Tuple[Var, Var]:
    r"""
    Support Vector Machine classifier

    Parameters
    ==========
    X
        Type T1.
        Data to be classified.
    classlabels_ints
        Attribute.
        Class labels if using integer labels.One and only one of the
        'classlabels\_\*' attributes must be defined.
    classlabels_strings
        Attribute.
        Class labels if using string labels.One and only one of the
        'classlabels\_\*' attributes must be defined.
    coefficients
        Attribute.

    kernel_params
        Attribute.
        List of 3 elements containing gamma, coef0, and degree, in that order.
        Zero if unused for the kernel.
    kernel_type
        Attribute.
        The kernel type, one of 'LINEAR,' 'POLY,' 'RBF,' 'SIGMOID'.
    post_transform
        Attribute.
        Indicates the transform to apply to the score. One of 'NONE,' 'SOFTMAX,'
        'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'
    prob_a
        Attribute.
        First set of probability coefficients.
    prob_b
        Attribute.
        Second set of probability coefficients. This array must be same size as
        prob_a.If these are provided then output Z are probability estimates,
        otherwise they are raw scores.
    rho
        Attribute.

    support_vectors
        Attribute.

    vectors_per_class
        Attribute.


    Returns
    =======
    Y : Var
        Type T2.
        Classification outputs (one class per example).
    Z : Var
        Type tensor(float).
        Class scores (one per class per example), if prob_a and prob_b are
        provided they are probabilities for each class, otherwise they are raw
        scores.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::SVMClassifier``.

    Type constraints:
     - T1: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
     - T2: `tensor(int64)`, `tensor(string)`
    """
    return _SVMClassifier(
        _SVMClassifier.Attributes(
            classlabels_ints=AttrInt64s.maybe(
                classlabels_ints, name="classlabels_ints"
            ),
            classlabels_strings=AttrStrings.maybe(
                classlabels_strings, name="classlabels_strings"
            ),
            coefficients=AttrFloat32s.maybe(coefficients, name="coefficients"),
            kernel_params=AttrFloat32s.maybe(kernel_params, name="kernel_params"),
            kernel_type=AttrString(kernel_type, name="kernel_type"),
            post_transform=AttrString(post_transform, name="post_transform"),
            prob_a=AttrFloat32s.maybe(prob_a, name="prob_a"),
            prob_b=AttrFloat32s.maybe(prob_b, name="prob_b"),
            rho=AttrFloat32s.maybe(rho, name="rho"),
            support_vectors=AttrFloat32s.maybe(support_vectors, name="support_vectors"),
            vectors_per_class=AttrInt64s.maybe(
                vectors_per_class, name="vectors_per_class"
            ),
        ),
        _SVMClassifier.Inputs(
            X=X,
        ),
    ).outputs._unpack_to_any()


def svmregressor(
    X: Var,
    *,
    coefficients: Optional[Iterable[float]] = None,
    kernel_params: Optional[Iterable[float]] = None,
    kernel_type: str = "LINEAR",
    n_supports: int = 0,
    one_class: int = 0,
    post_transform: str = "NONE",
    rho: Optional[Iterable[float]] = None,
    support_vectors: Optional[Iterable[float]] = None,
) -> Var:
    r"""
    Support Vector Machine regression prediction and one-class SVM anomaly
    detection.

    Parameters
    ==========
    X
        Type T.
        Data to be regressed.
    coefficients
        Attribute.
        Support vector coefficients.
    kernel_params
        Attribute.
        List of 3 elements containing gamma, coef0, and degree, in that order.
        Zero if unused for the kernel.
    kernel_type
        Attribute.
        The kernel type, one of 'LINEAR,' 'POLY,' 'RBF,' 'SIGMOID'.
    n_supports
        Attribute.
        The number of support vectors.
    one_class
        Attribute.
        Flag indicating whether the regression is a one-class SVM or not.
    post_transform
        Attribute.
        Indicates the transform to apply to the score. One of 'NONE,' 'SOFTMAX,'
        'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT.'
    rho
        Attribute.

    support_vectors
        Attribute.
        Chosen support vectors

    Returns
    =======
    Y : Var
        Type tensor(float).
        Regression outputs (one score per target per example).

    Notes
    =====
    Signature: ``ai.onnx.ml@1::SVMRegressor``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _SVMRegressor(
        _SVMRegressor.Attributes(
            coefficients=AttrFloat32s.maybe(coefficients, name="coefficients"),
            kernel_params=AttrFloat32s.maybe(kernel_params, name="kernel_params"),
            kernel_type=AttrString(kernel_type, name="kernel_type"),
            n_supports=AttrInt64(n_supports, name="n_supports"),
            one_class=AttrInt64(one_class, name="one_class"),
            post_transform=AttrString(post_transform, name="post_transform"),
            rho=AttrFloat32s.maybe(rho, name="rho"),
            support_vectors=AttrFloat32s.maybe(support_vectors, name="support_vectors"),
        ),
        _SVMRegressor.Inputs(
            X=X,
        ),
    ).outputs.Y


def scaler(
    X: Var,
    *,
    offset: Optional[Iterable[float]] = None,
    scale: Optional[Iterable[float]] = None,
) -> Var:
    r"""
    Rescale input data, for example to standardize features by removing the
    mean and scaling to unit variance.

    Parameters
    ==========
    X
        Type T.
        Data to be scaled.
    offset
        Attribute.
        First, offset by this.Can be length of features in an [N,F] tensor or
        length 1, in which case it applies to all features, regardless of
        dimension count.
    scale
        Attribute.
        Second, multiply by this.Can be length of features in an [N,F] tensor or
        length 1, in which case it applies to all features, regardless of
        dimension count.Must be same length as 'offset'

    Returns
    =======
    Y : Var
        Type tensor(float).
        Scaled output data.

    Notes
    =====
    Signature: ``ai.onnx.ml@1::Scaler``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _Scaler(
        _Scaler.Attributes(
            offset=AttrFloat32s.maybe(offset, name="offset"),
            scale=AttrFloat32s.maybe(scale, name="scale"),
        ),
        _Scaler.Inputs(
            X=X,
        ),
    ).outputs.Y


def tree_ensemble_classifier(
    X: Var,
    *,
    base_values: Optional[Iterable[float]] = None,
    base_values_as_tensor: Optional[np.ndarray] = None,
    class_ids: Optional[Iterable[int]] = None,
    class_nodeids: Optional[Iterable[int]] = None,
    class_treeids: Optional[Iterable[int]] = None,
    class_weights: Optional[Iterable[float]] = None,
    class_weights_as_tensor: Optional[np.ndarray] = None,
    classlabels_int64s: Optional[Iterable[int]] = None,
    classlabels_strings: Optional[Iterable[str]] = None,
    nodes_falsenodeids: Optional[Iterable[int]] = None,
    nodes_featureids: Optional[Iterable[int]] = None,
    nodes_hitrates: Optional[Iterable[float]] = None,
    nodes_hitrates_as_tensor: Optional[np.ndarray] = None,
    nodes_missing_value_tracks_true: Optional[Iterable[int]] = None,
    nodes_modes: Optional[Iterable[str]] = None,
    nodes_nodeids: Optional[Iterable[int]] = None,
    nodes_treeids: Optional[Iterable[int]] = None,
    nodes_truenodeids: Optional[Iterable[int]] = None,
    nodes_values: Optional[Iterable[float]] = None,
    nodes_values_as_tensor: Optional[np.ndarray] = None,
    post_transform: str = "NONE",
) -> Tuple[Var, Var]:
    r"""
    Tree Ensemble classifier. Returns the top class for each of N inputs.
    The attributes named 'nodes_X' form a sequence of tuples, associated by
    index into the sequences, which must all be of equal length. These
    tuples define the nodes. Similarly, all fields prefixed with 'class\_'
    are tuples of votes at the leaves. A leaf may have multiple votes, where
    each vote is weighted by the associated class_weights index. One and
    only one of classlabels_strings or classlabels_int64s will be defined.
    The class_ids are indices into this list. All fields ending with
    \_as_tensor can be used instead of the same parameter without the suffix
    if the element type is double and not float.

    Parameters
    ==========
    X
        Type T1.
        Input of shape [N,F]
    base_values
        Attribute.
        Base values for classification, added to final class score; the size
        must be the same as the classes or can be left unassigned (assumed 0)
    base_values_as_tensor
        Attribute.
        Base values for classification, added to final class score; the size
        must be the same as the classes or can be left unassigned (assumed 0)
    class_ids
        Attribute.
        The index of the class list that each weight is for.
    class_nodeids
        Attribute.
        node id that this weight is for.
    class_treeids
        Attribute.
        The id of the tree that this node is in.
    class_weights
        Attribute.
        The weight for the class in class_id.
    class_weights_as_tensor
        Attribute.
        The weight for the class in class_id.
    classlabels_int64s
        Attribute.
        Class labels if using integer labels.One and only one of the
        'classlabels\_\*' attributes must be defined.
    classlabels_strings
        Attribute.
        Class labels if using string labels.One and only one of the
        'classlabels\_\*' attributes must be defined.
    nodes_falsenodeids
        Attribute.
        Child node if expression is false.
    nodes_featureids
        Attribute.
        Feature id for each node.
    nodes_hitrates
        Attribute.
        Popularity of each node, used for performance and may be omitted.
    nodes_hitrates_as_tensor
        Attribute.
        Popularity of each node, used for performance and may be omitted.
    nodes_missing_value_tracks_true
        Attribute.
        For each node, define what to do in the presence of a missing value: if
        a value is missing (NaN), use the 'true' or 'false' branch based on the
        value in this array.This attribute may be left undefined, and the
        default value is false (0) for all nodes.
    nodes_modes
        Attribute.
        The node kind, that is, the comparison to make at the node. There is no
        comparison to make at a leaf node.One of 'BRANCH_LEQ', 'BRANCH_LT',
        'BRANCH_GTE', 'BRANCH_GT', 'BRANCH_EQ', 'BRANCH_NEQ', 'LEAF'
    nodes_nodeids
        Attribute.
        Node id for each node. Ids may restart at zero for each tree, but it not
        required to.
    nodes_treeids
        Attribute.
        Tree id for each node.
    nodes_truenodeids
        Attribute.
        Child node if expression is true.
    nodes_values
        Attribute.
        Thresholds to do the splitting on for each node.
    nodes_values_as_tensor
        Attribute.
        Thresholds to do the splitting on for each node.
    post_transform
        Attribute.
        Indicates the transform to apply to the score. One of 'NONE,' 'SOFTMAX,'
        'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT.'

    Returns
    =======
    Y : Var
        Type T2.
        N, Top class for each point
    Z : Var
        Type tensor(float).
        The class score for each class, for each point, a tensor of shape [N,E].

    Notes
    =====
    Signature: ``ai.onnx.ml@3::TreeEnsembleClassifier``.

    Type constraints:
     - T1: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
     - T2: `tensor(int64)`, `tensor(string)`
    """
    return _TreeEnsembleClassifier(
        _TreeEnsembleClassifier.Attributes(
            base_values=AttrFloat32s.maybe(base_values, name="base_values"),
            base_values_as_tensor=AttrTensor.maybe(
                base_values_as_tensor, name="base_values_as_tensor"
            ),
            class_ids=AttrInt64s.maybe(class_ids, name="class_ids"),
            class_nodeids=AttrInt64s.maybe(class_nodeids, name="class_nodeids"),
            class_treeids=AttrInt64s.maybe(class_treeids, name="class_treeids"),
            class_weights=AttrFloat32s.maybe(class_weights, name="class_weights"),
            class_weights_as_tensor=AttrTensor.maybe(
                class_weights_as_tensor, name="class_weights_as_tensor"
            ),
            classlabels_int64s=AttrInt64s.maybe(
                classlabels_int64s, name="classlabels_int64s"
            ),
            classlabels_strings=AttrStrings.maybe(
                classlabels_strings, name="classlabels_strings"
            ),
            nodes_falsenodeids=AttrInt64s.maybe(
                nodes_falsenodeids, name="nodes_falsenodeids"
            ),
            nodes_featureids=AttrInt64s.maybe(
                nodes_featureids, name="nodes_featureids"
            ),
            nodes_hitrates=AttrFloat32s.maybe(nodes_hitrates, name="nodes_hitrates"),
            nodes_hitrates_as_tensor=AttrTensor.maybe(
                nodes_hitrates_as_tensor, name="nodes_hitrates_as_tensor"
            ),
            nodes_missing_value_tracks_true=AttrInt64s.maybe(
                nodes_missing_value_tracks_true, name="nodes_missing_value_tracks_true"
            ),
            nodes_modes=AttrStrings.maybe(nodes_modes, name="nodes_modes"),
            nodes_nodeids=AttrInt64s.maybe(nodes_nodeids, name="nodes_nodeids"),
            nodes_treeids=AttrInt64s.maybe(nodes_treeids, name="nodes_treeids"),
            nodes_truenodeids=AttrInt64s.maybe(
                nodes_truenodeids, name="nodes_truenodeids"
            ),
            nodes_values=AttrFloat32s.maybe(nodes_values, name="nodes_values"),
            nodes_values_as_tensor=AttrTensor.maybe(
                nodes_values_as_tensor, name="nodes_values_as_tensor"
            ),
            post_transform=AttrString(post_transform, name="post_transform"),
        ),
        _TreeEnsembleClassifier.Inputs(
            X=X,
        ),
    ).outputs._unpack_to_any()


def tree_ensemble_regressor(
    X: Var,
    *,
    aggregate_function: str = "SUM",
    base_values: Optional[Iterable[float]] = None,
    base_values_as_tensor: Optional[np.ndarray] = None,
    n_targets: Optional[int] = None,
    nodes_falsenodeids: Optional[Iterable[int]] = None,
    nodes_featureids: Optional[Iterable[int]] = None,
    nodes_hitrates: Optional[Iterable[float]] = None,
    nodes_hitrates_as_tensor: Optional[np.ndarray] = None,
    nodes_missing_value_tracks_true: Optional[Iterable[int]] = None,
    nodes_modes: Optional[Iterable[str]] = None,
    nodes_nodeids: Optional[Iterable[int]] = None,
    nodes_treeids: Optional[Iterable[int]] = None,
    nodes_truenodeids: Optional[Iterable[int]] = None,
    nodes_values: Optional[Iterable[float]] = None,
    nodes_values_as_tensor: Optional[np.ndarray] = None,
    post_transform: str = "NONE",
    target_ids: Optional[Iterable[int]] = None,
    target_nodeids: Optional[Iterable[int]] = None,
    target_treeids: Optional[Iterable[int]] = None,
    target_weights: Optional[Iterable[float]] = None,
    target_weights_as_tensor: Optional[np.ndarray] = None,
) -> Var:
    r"""
    Tree Ensemble regressor. Returns the regressed values for each input in
    N. All args with nodes\_ are fields of a tuple of tree nodes, and it is
    assumed they are the same length, and an index i will decode the tuple
    across these inputs. Each node id can appear only once for each tree id.
    All fields prefixed with target\_ are tuples of votes at the leaves. A
    leaf may have multiple votes, where each vote is weighted by the
    associated target_weights index. All fields ending with \_as_tensor can
    be used instead of the same parameter without the suffix if the element
    type is double and not float. All trees must have their node ids start
    at 0 and increment by 1. Mode enum is BRANCH_LEQ, BRANCH_LT, BRANCH_GTE,
    BRANCH_GT, BRANCH_EQ, BRANCH_NEQ, LEAF

    Parameters
    ==========
    X
        Type T.
        Input of shape [N,F]
    aggregate_function
        Attribute.
        Defines how to aggregate leaf values within a target. One of 'AVERAGE,'
        'SUM,' 'MIN,' 'MAX.'
    base_values
        Attribute.
        Base values for classification, added to final class score; the size
        must be the same as the classes or can be left unassigned (assumed 0)
    base_values_as_tensor
        Attribute.
        Base values for classification, added to final class score; the size
        must be the same as the classes or can be left unassigned (assumed 0)
    n_targets
        Attribute.
        The total number of targets.
    nodes_falsenodeids
        Attribute.
        Child node if expression is false
    nodes_featureids
        Attribute.
        Feature id for each node.
    nodes_hitrates
        Attribute.
        Popularity of each node, used for performance and may be omitted.
    nodes_hitrates_as_tensor
        Attribute.
        Popularity of each node, used for performance and may be omitted.
    nodes_missing_value_tracks_true
        Attribute.
        For each node, define what to do in the presence of a NaN: use the
        'true' (if the attribute value is 1) or 'false' (if the attribute value
        is 0) branch based on the value in this array.This attribute may be left
        undefined and the default value is false (0) for all nodes.
    nodes_modes
        Attribute.
        The node kind, that is, the comparison to make at the node. There is no
        comparison to make at a leaf node.One of 'BRANCH_LEQ', 'BRANCH_LT',
        'BRANCH_GTE', 'BRANCH_GT', 'BRANCH_EQ', 'BRANCH_NEQ', 'LEAF'
    nodes_nodeids
        Attribute.
        Node id for each node. Node ids must restart at zero for each tree and
        increase sequentially.
    nodes_treeids
        Attribute.
        Tree id for each node.
    nodes_truenodeids
        Attribute.
        Child node if expression is true
    nodes_values
        Attribute.
        Thresholds to do the splitting on for each node.
    nodes_values_as_tensor
        Attribute.
        Thresholds to do the splitting on for each node.
    post_transform
        Attribute.
        Indicates the transform to apply to the score. One of 'NONE,' 'SOFTMAX,'
        'LOGISTIC,' 'SOFTMAX_ZERO,' or 'PROBIT'
    target_ids
        Attribute.
        The index of the target that each weight is for
    target_nodeids
        Attribute.
        The node id of each weight
    target_treeids
        Attribute.
        The id of the tree that each node is in.
    target_weights
        Attribute.
        The weight for each target
    target_weights_as_tensor
        Attribute.
        The weight for each target

    Returns
    =======
    Y : Var
        Type tensor(float).
        N classes

    Notes
    =====
    Signature: ``ai.onnx.ml@3::TreeEnsembleRegressor``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(int32)`, `tensor(int64)`
    """
    return _TreeEnsembleRegressor(
        _TreeEnsembleRegressor.Attributes(
            aggregate_function=AttrString(
                aggregate_function, name="aggregate_function"
            ),
            base_values=AttrFloat32s.maybe(base_values, name="base_values"),
            base_values_as_tensor=AttrTensor.maybe(
                base_values_as_tensor, name="base_values_as_tensor"
            ),
            n_targets=AttrInt64.maybe(n_targets, name="n_targets"),
            nodes_falsenodeids=AttrInt64s.maybe(
                nodes_falsenodeids, name="nodes_falsenodeids"
            ),
            nodes_featureids=AttrInt64s.maybe(
                nodes_featureids, name="nodes_featureids"
            ),
            nodes_hitrates=AttrFloat32s.maybe(nodes_hitrates, name="nodes_hitrates"),
            nodes_hitrates_as_tensor=AttrTensor.maybe(
                nodes_hitrates_as_tensor, name="nodes_hitrates_as_tensor"
            ),
            nodes_missing_value_tracks_true=AttrInt64s.maybe(
                nodes_missing_value_tracks_true, name="nodes_missing_value_tracks_true"
            ),
            nodes_modes=AttrStrings.maybe(nodes_modes, name="nodes_modes"),
            nodes_nodeids=AttrInt64s.maybe(nodes_nodeids, name="nodes_nodeids"),
            nodes_treeids=AttrInt64s.maybe(nodes_treeids, name="nodes_treeids"),
            nodes_truenodeids=AttrInt64s.maybe(
                nodes_truenodeids, name="nodes_truenodeids"
            ),
            nodes_values=AttrFloat32s.maybe(nodes_values, name="nodes_values"),
            nodes_values_as_tensor=AttrTensor.maybe(
                nodes_values_as_tensor, name="nodes_values_as_tensor"
            ),
            post_transform=AttrString(post_transform, name="post_transform"),
            target_ids=AttrInt64s.maybe(target_ids, name="target_ids"),
            target_nodeids=AttrInt64s.maybe(target_nodeids, name="target_nodeids"),
            target_treeids=AttrInt64s.maybe(target_treeids, name="target_treeids"),
            target_weights=AttrFloat32s.maybe(target_weights, name="target_weights"),
            target_weights_as_tensor=AttrTensor.maybe(
                target_weights_as_tensor, name="target_weights_as_tensor"
            ),
        ),
        _TreeEnsembleRegressor.Inputs(
            X=X,
        ),
    ).outputs.Y


def zip_map(
    X: Var,
    *,
    classlabels_int64s: Optional[Iterable[int]] = None,
    classlabels_strings: Optional[Iterable[str]] = None,
) -> Var:
    r"""
    Creates a map from the input and the attributes. The values are provided
    by the input tensor, while the keys are specified by the attributes.
    Must provide keys in either classlabels_strings or classlabels_int64s
    (but not both). The columns of the tensor correspond one-by-one to the
    keys specified by the attributes. There must be as many columns as keys.

    Parameters
    ==========
    X
        Type tensor(float).
        The input values
    classlabels_int64s
        Attribute.
        The keys when using int keys.One and only one of the 'classlabels\_\*'
        attributes must be defined.
    classlabels_strings
        Attribute.
        The keys when using string keys.One and only one of the
        'classlabels\_\*' attributes must be defined.

    Returns
    =======
    Z : Var
        Type T.
        The output map

    Notes
    =====
    Signature: ``ai.onnx.ml@1::ZipMap``.

    Type constraints:
     - T: `seq(map(int64,tensor(float)))`, `seq(map(string,tensor(float)))`
    """
    return _ZipMap(
        _ZipMap.Attributes(
            classlabels_int64s=AttrInt64s.maybe(
                classlabels_int64s, name="classlabels_int64s"
            ),
            classlabels_strings=AttrStrings.maybe(
                classlabels_strings, name="classlabels_strings"
            ),
        ),
        _ZipMap.Inputs(
            X=X,
        ),
    ).outputs.Z


_OPERATORS = {
    "ArrayFeatureExtractor": _ArrayFeatureExtractor,
    "Binarizer": _Binarizer,
    "CastMap": _CastMap,
    "CategoryMapper": _CategoryMapper,
    "DictVectorizer": _DictVectorizer,
    "FeatureVectorizer": _FeatureVectorizer,
    "Imputer": _Imputer,
    "LabelEncoder": _LabelEncoder,
    "LinearClassifier": _LinearClassifier,
    "LinearRegressor": _LinearRegressor,
    "Normalizer": _Normalizer,
    "OneHotEncoder": _OneHotEncoder,
    "SVMClassifier": _SVMClassifier,
    "SVMRegressor": _SVMRegressor,
    "Scaler": _Scaler,
    "TreeEnsembleClassifier": _TreeEnsembleClassifier,
    "TreeEnsembleRegressor": _TreeEnsembleRegressor,
    "ZipMap": _ZipMap,
}

_CONSTRUCTORS = {
    "ArrayFeatureExtractor": array_feature_extractor,
    "Binarizer": binarizer,
    "CastMap": cast_map,
    "CategoryMapper": category_mapper,
    "DictVectorizer": dict_vectorizer,
    "FeatureVectorizer": feature_vectorizer,
    "Imputer": imputer,
    "LabelEncoder": label_encoder,
    "LinearClassifier": linear_classifier,
    "LinearRegressor": linear_regressor,
    "Normalizer": normalizer,
    "OneHotEncoder": one_hot_encoder,
    "SVMClassifier": svmclassifier,
    "SVMRegressor": svmregressor,
    "Scaler": scaler,
    "TreeEnsembleClassifier": tree_ensemble_classifier,
    "TreeEnsembleRegressor": tree_ensemble_regressor,
    "ZipMap": zip_map,
}

__all__ = [fun.__name__ for fun in _CONSTRUCTORS.values()]
